#include <cassert>
#include <climits>
#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <utility>     // std::pair, make_pair
#include <algorithm>
#include <map>
#include <set>
#include <vector>

#ifdef USE_CLOCK
#include <ctime>
#endif

#define MAX_PIECES 30

#define VERSION "0.05"

/*
 * tiling.C: determine how a number of shapes can cover another shape.
 * Run without arguments for usage information.
 * 
 * Copyright (C) 2016-2018, David Moews.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */


/*                              C O U N T S                    */

typedef unsigned long long Count;

static Count solns, nodes, checks;

/*                  E N D       C O U N T S                    */


/*                        C L O C K           (OS- and CPU-dependent)         */

#ifdef USE_CLOCK

Count now_rdtsc()
{
  unsigned int a, b;
  asm volatile("rdtsc" : "=a" (a), "=d" (b));
  return (((unsigned long long)b) << 32) | ((unsigned long long)a);
}

Count now_realtime()
{
  struct timespec t;
  clock_gettime(CLOCK_REALTIME, &t);
  return ((unsigned long long)t.tv_sec) * 1000000000ul + ((unsigned long long)t.tv_nsec);
}

Count now_cpuprocess()
{
  struct timespec t;
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t);
  return ((unsigned long long)t.tv_sec) * 1000000000ul + ((unsigned long long)t.tv_nsec);
}

struct Times
{
  Count rdtsc, realtime, cpuprocess;

  Times() : rdtsc(0), realtime(0), cpuprocess(0) { }
  Times(Count ts, Count rt, Count pr) 
     : rdtsc(ts), realtime(rt), cpuprocess(pr) { }
   
};

void get_times(Times *p)
{
  p->rdtsc = now_rdtsc();
  p->realtime = now_realtime();
  p->cpuprocess = now_cpuprocess();
}

void add_times(const Times &a, const Times &b, Times *p_c)
{
  assert(b.rdtsc >= a.rdtsc);
  p_c->rdtsc += b.rdtsc - a.rdtsc;
  assert(b.realtime >= a.realtime);
  p_c->realtime += b.realtime - a.realtime;
  assert(b.cpuprocess >= a.cpuprocess);
  p_c->cpuprocess += b.cpuprocess - a.cpuprocess;
}

#else

struct Times { };

void get_times(Times *)
{
}

void add_times(const Times &, const Times &, Times *)
{
}

#endif

static Times search_start;
static Times search_end;
static Times search_time;
static Times setup_time;

/*                 E N D       C L O C K                    */

/*       S H A P E S         (and symmetries)                */

enum SymmetryGroup { SG_none, SG_trans, SG_rot, SG_rot_refl };

class Shape
{
  std::vector<int> bb_lo;
  std::vector<int> bb_hi;
  std::vector<int> coords;
  std::vector<int> coord_sums;
  bool nonempty_flag;

public:
  // Creates empty shape with n_dims dimensions.
  Shape(int n_dims = 0)
     : bb_lo(std::vector<int>(n_dims, INT_MAX)), 
       bb_hi(std::vector<int>(n_dims, INT_MIN)), 
       coords(),
       coord_sums(std::vector<int>(n_dims, 0)),
       nonempty_flag(false) { }

  // Creates box with given dimensions (assumed to be nonnegative.)
  Shape(const std::vector<int> &);

  // Sets to empty shape with n_dims dimensions.
  void make_empty(int n_dims = 0)
  {
    bb_lo.assign(n_dims, INT_MAX);
    bb_hi.assign(n_dims, INT_MIN);
    coords.clear();
    coord_sums.assign(n_dims, 0);
    nonempty_flag = false;
  }

  int get_n_dims() const { return bb_lo.size(); }

  // Get pointers to bounding box corner coordinates.
  const int *get_bb_lo() const { return bb_lo.empty() ? (const int *)NULL : &bb_lo[0]; }
  const int *get_bb_hi() const { return bb_hi.empty() ? (const int *)NULL : &bb_hi[0]; }

  std::size_t n_cubes() const 
  {
    if (bb_lo.empty())
       return nonempty_flag ? 1 : 0;

    assert((coords.size() % get_n_dims()) == 0);
    return coords.size() / get_n_dims();
  }

  // Get pointer to coordinates of ith cube.
  const int *get_cube(std::size_t i) const
  {
    assert(i < n_cubes());

    if (coords.empty())
       return NULL;
    else
       return &coords[bb_lo.size() * i];
  }

  // Return whether cube with given coordinates (first argument) is present.
  // If found and second argument nonnull, put index into *(second argument).
  // Otherwise, if second argument nonnull, put index of where to add this
  // cube into *(second argument).
  bool find_cube(const int *, std::size_t *) const;

  void normalize_by_translation();

  void translate_by(const int *);

  // Returns true if cube successfully added, false if it was already there.
  bool add_cube(const int *);

  void set_to_symmetry_image(SymmetryGroup, int, const Shape &);

  void set_to_normalized_symmetry_image(SymmetryGroup, int, const Shape &);

  // Returns index of symmetry used to get minimal normalized symmetry image.
  int set_to_minimal_normalized_symmetry_image(SymmetryGroup, const Shape &);

  void pad_dimensions(int);

  friend int cmp(const Shape &, const Shape &);
  friend bool operator ==(const Shape &, const Shape &);
};


/* Construct filled box with corner at the origin. */
Shape::Shape(const std::vector<int> &dims) 
  : bb_lo(dims.size(), 0), bb_hi(dims.size(), 0), 
    coords(), coord_sums(dims.size(), 0), nonempty_flag(true)
{
  int n_dims = dims.size();
  std::size_t volume = 1;

  for (int i = 0; i < n_dims; i++)
  {
    if (dims[i] == 0)
    {
      // Empty.
      bb_lo.assign(dims.size(), INT_MAX);
      bb_hi.assign(dims.size(), INT_MIN);
      nonempty_flag = false;
      return;
    }

    assert(dims[i] > 0);
    bb_hi[i] = dims[i] - 1;

    volume *= dims[i]; 
  }

  coords.resize(volume * n_dims);

  std::vector<int> x(n_dims, 0);

  for (std::size_t i = 0; i < volume * n_dims; i += n_dims)
  {
    for (int j = 0; j < n_dims; j++)
        coord_sums[j] += (coords[i + j] = x[j]);

    for (int j = n_dims - 1; j >= 0; j--)
    {
      if (++x[j] != dims[j])
         break;
      x[j] = 0;
    }
  }
}

void Shape::normalize_by_translation()
{
  if (coords.empty())
     return;

  int i_chk;

  int n_dims = bb_lo.size();

  for (i_chk = 0; i_chk < n_dims; i_chk++)
  if (bb_lo[i_chk] != 0)
      break;

  if (i_chk == n_dims)
     return;

  for (std::size_t i = 0; i < coords.size(); i += n_dims)
  for (int j = i_chk; j < n_dims; j++)
      coords[i + j] -= bb_lo[j];

  int n_c = n_cubes();

  for (int j = i_chk; j < n_dims; j++)
  {
    coord_sums[j] -= n_c * bb_lo[j];
    bb_hi[j] -= bb_lo[j];
    bb_lo[j] = 0;
  }
}

void Shape::translate_by(const int *n)
{
  if (coords.empty())
     return;

  int n_dims = bb_lo.size();

  for (std::size_t i = 0; i < coords.size(); i += n_dims)
  for (int j = 0; j < n_dims; j++)
      coords[i + j] += n[j];

  int n_c = n_cubes();

  for (int j = 0; j < n_dims; j++)
  {
    coord_sums[j] += n_c * n[j];
    bb_lo[j] += n[j];
    bb_hi[j] += n[j];
  }
}

bool Shape::find_cube(const int *cube_coords, std::size_t *p_n) const
{
  if (coords.empty())
  {
    if (p_n != NULL)
       *p_n = 0;

    return nonempty_flag;
  }

  int n_dimensions = get_n_dims();
  std::size_t lo = 0;  
  std::size_t hi = coords.size() / n_dimensions - 1;

  while (lo <= hi)
  {
    /*
     * Loop invariant: 
     * new_cube > cubes 0, ..., lo-1
     * new_cube < cubes hi+1, ..., n-1
     */
    std::size_t med = (lo + hi) / 2;
    std::size_t med_idx = med * n_dimensions;
    int j;

    for (j = 0; j < n_dimensions; j++)
    if (cube_coords[j] != coords[med_idx + j])
       break;

    if (j == n_dimensions)
    {
      if (p_n != NULL)
         *p_n = med;

      return true;
    }

    if (cube_coords[j] < coords[med_idx + j])
    {
      if (med == 0)
         break;

      hi = med - 1;
    } else
      lo = med + 1;
  }

  if (p_n != NULL)
     *p_n = lo;

  return false;
}

bool Shape::add_cube(const int *cube_coords)
{
  int n_dims = bb_lo.size();

  if (coords.empty())
  {
    if (nonempty_flag)
    {
      assert(n_dims == 0);
      return false;
    }

    nonempty_flag = true;
    coords.assign(cube_coords, cube_coords + n_dims);
    coord_sums.assign(cube_coords, cube_coords + n_dims);
    bb_lo.assign(cube_coords, cube_coords + n_dims);
    bb_hi.assign(cube_coords, cube_coords + n_dims);

    return true;
  }

  std::size_t idx;

  if (find_cube(cube_coords, &idx))
     return false;   /* Cube already present */
   
  coords.insert(coords.begin() + idx * n_dims, 
                cube_coords, cube_coords + n_dims);

  for (int i = 0; i < n_dims; i++)
  {
    coord_sums[i] += cube_coords[i];
    if (cube_coords[i] < bb_lo[i])
       bb_lo[i] = cube_coords[i];
    if (cube_coords[i] > bb_hi[i])
       bb_hi[i] = cube_coords[i];
  }

  return true;
}

bool operator ==(const Shape &a, const Shape &b)
{
  if ((a.nonempty_flag != b.nonempty_flag) ||
      (a.bb_lo != b.bb_lo) ||
      (a.bb_hi != b.bb_hi) ||
      (a.coord_sums != b.coord_sums) ||
      (a.coords != b.coords))
     return false;

  return true;
}

int cmp(const Shape &a, const Shape &b)
{
  if (a.get_n_dims() < b.get_n_dims())
     return -1;

  if (a.get_n_dims() > b.get_n_dims())
     return 1;

  if (a.nonempty_flag < b.nonempty_flag)
     return -1;

  if (a.nonempty_flag > b.nonempty_flag)
     return 1;

  if (a.coords.size() < b.coords.size())
     return -1;

  if (a.coords.size() > b.coords.size())
     return 1;

  for (int i = 0; i < a.get_n_dims(); i++)
  {
    if (a.bb_lo[i] < b.bb_lo[i])
       return -1;
    if (a.bb_lo[i] > b.bb_lo[i])
       return 1;
    if (a.bb_hi[i] < b.bb_hi[i])
       return -1;
    if (a.bb_hi[i] > b.bb_hi[i])
       return 1;
    if (a.coord_sums[i] < b.coord_sums[i])
       return -1;
    if (a.coord_sums[i] > b.coord_sums[i])
       return 1;
  }

  for (std::size_t i = 0; i < a.coords.size(); i++)
  {
    if (a.coords[i] < b.coords[i])
       return -1;
    if (a.coords[i] > b.coords[i])
       return 1;
  }

  return 0;  /* Equal */
}

inline bool operator <(const Shape &a, const Shape &b)
{
  return cmp(a, b) < 0;
}

static int cc_n_dims = 0;

int compare_coords(const void *a, const void *b)
{
  const int *c_a = (const int *)a;
  const int *c_b = (const int *)b;

  for (int i = 0; i < cc_n_dims; i++)
  {
    if (c_a[i] < c_b[i])
       return -1;

    if (c_a[i] > c_b[i])
       return 1;
  }

  return 0;
}

int factorial(int n)
{
  int rv = 1;

  assert(n >= 0);
  for (int i = 1; i <= n; i++)
      rv *= i; 

  return rv;
}


int n_symmetries_in(SymmetryGroup sg, int n_dims)
{
  switch (sg)
  {
    case SG_none:
    case SG_trans:
         return 1;

    case SG_rot:
         if (n_dims == 0)
            return 1;
         else
            return factorial(n_dims) << (n_dims - 1);

    case SG_rot_refl:
         return factorial(n_dims) << n_dims;
  }
  /*NOTREACHED*/
  return 0;
}

void get_symmetry_operation(SymmetryGroup sg, int n_dims, int s_no, int *perm, int *sign)
{
  assert(s_no >= 0);

  if ((sg == SG_none) || (sg == SG_trans))
  {
    assert(s_no == 0);

    for (int i = 0; i < n_dims; i++)
    {
      perm[i] = i;
      sign[i] = 1;
    }
    return;
  }

  if (n_dims == 0)
  {
    assert(s_no == 0);
    return;
  }

  for (int i = 0; i < n_dims; i++)
      perm[i] = i;

  int sgn = 1;

  for (int i = 0; i < n_dims; i++)
  {
    int j = s_no % (n_dims - i);
    s_no /= n_dims - i;
    if (j != 0)
       sgn = -sgn;    
    std::swap(perm[i], perm[i + j]);
  }
  for (int i = 1; i < n_dims; i++)
  {
    if ((s_no & 1) != 0)
    {
      sign[i] = -1;
      sgn = -sgn;
    } else
      sign[i] = 1;
    s_no >>= 1;
  }

  if (s_no == 0)
     sign[0] = sgn;
  else
  {
    assert((sg == SG_rot_refl) && (s_no == 1));
    sign[0] = -sgn;
  }
}

/* Sets shape to symmetry image of s. */
void Shape::set_to_symmetry_image(SymmetryGroup sg, int s_no, const Shape &s) 
{
  int n_dims = s.get_n_dims();

  if (!s.nonempty_flag)
  {
    make_empty(n_dims);
    return;
  }

  nonempty_flag = true;
  bb_lo.resize(n_dims);
  bb_hi.resize(n_dims);
  coord_sums.resize(n_dims);
  coords.resize(s.coords.size());

  if (n_dims == 0)
     return;

  std::vector<int> perm(n_dims, 0);
  std::vector<int> sign(n_dims, 0);

  get_symmetry_operation(sg, n_dims, s_no, &perm[0], &sign[0]);

  for (int i = 0; i < n_dims; i++)
  {
    coord_sums[i] = sign[i] * s.coord_sums[perm[i]];

    if (sign[i] > 0)
    {
      bb_lo[i] = s.bb_lo[perm[i]];
      bb_hi[i] = s.bb_hi[perm[i]];
    } else
    {
      bb_lo[i] = -s.bb_hi[perm[i]];
      bb_hi[i] = -s.bb_lo[perm[i]];
    }
  }

  for (std::size_t i = 0; i < coords.size(); i += n_dims)
  for (int j = 0; j < n_dims; j++)
      coords[i + j] = sign[j] * s.coords[i + perm[j]];

  assert(!coords.empty());
  cc_n_dims = n_dims;
  std::qsort(&coords[0], n_cubes(), n_dims * sizeof(int), compare_coords);  
}

/* Sets shape to translation-normalized symmetry image of s. */
void Shape::set_to_normalized_symmetry_image(SymmetryGroup sg, int s_no, const Shape &s) 
{
  int n_dims = s.get_n_dims();

  if (!s.nonempty_flag)
  {
    make_empty(n_dims);
    return;
  }

  nonempty_flag = true;
  bb_lo.assign(n_dims, 0);
  bb_hi.resize(n_dims);
  coord_sums.resize(n_dims);
  coords.resize(s.coords.size());

  if (n_dims == 0)
     return;

  std::vector<int> perm(n_dims, 0);
  std::vector<int> sign(n_dims, 0);

  get_symmetry_operation(sg, n_dims, s_no, &perm[0], &sign[0]);

  int n_c = s.n_cubes();

  for (int i = 0; i < n_dims; i++)
  {
    bb_hi[i] = s.bb_hi[perm[i]] - s.bb_lo[perm[i]];
    if (sign[i] > 0)
       coord_sums[i] = s.coord_sums[perm[i]] - n_c * s.bb_lo[perm[i]];
    else
       coord_sums[i] = n_c * s.bb_hi[perm[i]] - s.coord_sums[perm[i]];
  }

  for (std::size_t i = 0; i < coords.size(); i += n_dims)
  for (int j = 0; j < n_dims; j++)
  if (sign[j] > 0)
     coords[i + j] = s.coords[i + perm[j]] - s.bb_lo[perm[j]];
  else
     coords[i + j] = s.bb_hi[perm[j]] - s.coords[i + perm[j]];

  assert(!coords.empty());
  cc_n_dims = n_dims;
  std::qsort(&coords[0], n_c, n_dims * sizeof(int), compare_coords);  
}

/* Sets shape to minimal symmetry image of s, where the symmetry image is required to be */
/* translation-normalized if the group includes translations.  Returns number of symmetry used. */
int Shape::set_to_minimal_normalized_symmetry_image(SymmetryGroup sg, const Shape &s)
{
  if (sg == SG_none)
  {
    *this = s;
    return 0;
  }

  int n_dims = s.get_n_dims();

  if (!s.nonempty_flag)
  {
    make_empty(n_dims);
    return 0;
  }

  nonempty_flag = true;
  bb_lo.assign(n_dims, 0);
  bb_hi.assign(n_dims, INT_MAX);
  coord_sums.assign(n_dims, INT_MAX);
  coords.assign(s.coords.size(), INT_MAX);

  if (n_dims == 0)
     return 0;

  std::vector<int> perm(n_dims, 0);
  std::vector<int> sign(n_dims, 0);

  int n_symmetries = n_symmetries_in(sg, n_dims);
  int n_c = s.n_cubes();

  std::vector<int> temp_bb_hi(n_dims, 0);
  std::vector<int> temp_coord_sums(n_dims, 0);
  std::vector<bool> symmetry_ok(n_symmetries, false);

  int s_no_used = -1;

  // First pass: make bb_hi, coord_sums minimal.

  for (int s_no = 0; s_no < n_symmetries; s_no++)
  {
    get_symmetry_operation(sg, n_dims, s_no, &perm[0], &sign[0]);

    for (int i = 0; i < n_dims; i++)
    {
      temp_bb_hi[i] = s.bb_hi[perm[i]] - s.bb_lo[perm[i]];
      temp_coord_sums[i] = (
         (sign[i] > 0) ?
                   s.coord_sums[perm[i]] - n_c * s.bb_lo[perm[i]] :
                   n_c * s.bb_hi[perm[i]] - s.coord_sums[perm[i]] 
                           );
    }

    int i;

    for (i = 0; i < n_dims; i++)
    if ((temp_bb_hi[i] != bb_hi[i]) || (coord_sums[i] != temp_coord_sums[i]))
       break;

    if (i == n_dims)    // Same.
    {
      symmetry_ok[s_no] = true;
      continue;      
    }

    if ((temp_bb_hi[i] < bb_hi[i]) ||
        ((temp_bb_hi[i] == bb_hi[i]) && (temp_coord_sums[i] < coord_sums[i])))   // Less
    {
      bb_hi = temp_bb_hi;
      coord_sums = temp_coord_sums;
      s_no_used = s_no;
    }
  }

  // Second pass

  assert((s_no_used >= 0) && (s_no_used < n_symmetries));

  symmetry_ok[s_no_used] = true;

  std::vector<int> temp_coords(coords.size(), 0);

  int s_no_used_pass_2 = -1;

  for (int s_no = s_no_used; s_no < n_symmetries; s_no++)
  if (symmetry_ok[s_no])
  {
    get_symmetry_operation(sg, n_dims, s_no, &perm[0], &sign[0]);

#ifndef NO_DEBUG_ASSERTS
    // Debug asserts.
    for (int i = 0; i < n_dims; i++)
    {
      assert(bb_hi[i] == s.bb_hi[perm[i]] - s.bb_lo[perm[i]]);
      assert(coord_sums[i] ==
                          (
         (sign[i] > 0) ?
                   s.coord_sums[perm[i]] - n_c * s.bb_lo[perm[i]] :
                   n_c * s.bb_hi[perm[i]] - s.coord_sums[perm[i]] 
                          )
            );
    }
#endif

    for (std::size_t i = 0; i < coords.size(); i += n_dims)
    for (int j = 0; j < n_dims; j++)
    if (sign[j] > 0)
       temp_coords[i + j] = s.coords[i + perm[j]] - s.bb_lo[perm[j]];
    else
       temp_coords[i + j] = s.bb_hi[perm[j]] - s.coords[i + perm[j]];

    assert(!temp_coords.empty());
    cc_n_dims = n_dims;
    std::qsort(&temp_coords[0], n_c, n_dims * sizeof(int), compare_coords);  

    if (temp_coords < coords)   // Less
    {
      s_no_used_pass_2 = s_no;
      coords = temp_coords;
    }
  }
  assert((s_no_used_pass_2 >= 0) && (s_no_used_pass_2 < n_symmetries));

  return s_no_used_pass_2;
}

/* Pad shape to a higher number of dimensions. */
/* A padded shape minimal in its symmetry class (with the appropriately reduced */
/* symmetry group) is still minimal after padding. */
void Shape::pad_dimensions(int new_n_dims)
{
  int n_dims = bb_lo.size();

  if (new_n_dims == n_dims)
     return;

  assert(new_n_dims > n_dims);

  if (!nonempty_flag)
  {
    bb_lo.resize(new_n_dims, INT_MAX);
    bb_hi.resize(new_n_dims, INT_MIN);
    coord_sums.resize(new_n_dims, 0);
    return;
  }

  std::size_t n_c = n_cubes();

  bb_lo.insert(bb_lo.begin(), new_n_dims - n_dims, 0);
  bb_hi.insert(bb_hi.begin(), new_n_dims - n_dims, 0);
  coord_sums.insert(coord_sums.begin(), new_n_dims - n_dims, 0);

  if (coords.empty())
  {
    assert((n_dims == 0) && (n_c == 1));

    coords.resize(new_n_dims, 0);
    return;
  }

  assert(n_c > 0);

  coords.resize(n_c * new_n_dims);

  for (std::size_t i = n_c - 1; i != 0; i--)
  {
    for (int j = n_dims - 1; j >= 0; j--)
        coords[i * new_n_dims + j + new_n_dims - n_dims] = coords[i * n_dims + j];
    for (int j = new_n_dims - n_dims - 1; j >= 0; j--)
        coords[i * new_n_dims + j] = 0;
  }
  for (int j = n_dims - 1; j >= 0; j--)
      coords[j + new_n_dims - n_dims] = coords[j];
  for (int j = new_n_dims - n_dims - 1; j >= 0; j--)
      coords[j] = 0;
}

/* Rotations in higher dimensions can restrict to lower-dimensional reflections. */
SymmetryGroup restrict_group(SymmetryGroup sg, int big_dims, int small_dims)
{
  assert((big_dims >= small_dims) && (small_dims >= 0));

  if ((sg == SG_rot) && (small_dims < big_dims))
     return SG_rot_refl;

  return sg;
}

/*       E N D                   S H A P E S               */

/*           B I T M A S K S                             */

#define BITS_PER_BITMASK ((int)(8 * sizeof(BitmaskBaseType) * WORDS_PER_BITMASK))

typedef unsigned long long BitmaskBaseType;

#ifndef WORDS_PER_BITMASK
#define WORDS_PER_BITMASK 4
#endif

class Bitmask
{

#if WORDS_PER_BITMASK == 1

  BitmaskBaseType val;

  Bitmask(BitmaskBaseType v) : val(v) { }

public:

  /* Default Bitmask is zero (no bits set.) */
  Bitmask() : val(0) { }

#else

  /* Array bitmask. */

  BitmaskBaseType val[WORDS_PER_BITMASK];

public:

  /* Default Bitmask is zero (no bits set.) */
  Bitmask() 
  {
    for (int i = 0; i < WORDS_PER_BITMASK; i++)
        val[i] = 0;
  }

#endif

  friend Bitmask operator &(const Bitmask &, const Bitmask &);
  friend Bitmask operator |(const Bitmask &, const Bitmask &);
  friend Bitmask operator ~(const Bitmask &);

  friend bool is_zero(const Bitmask &);
  friend bool operator ==(const Bitmask &, const Bitmask &);
  friend bool operator !=(const Bitmask &, const Bitmask &);
  friend bool operator <(const Bitmask &, const Bitmask &);
  friend bool operator >(const Bitmask &, const Bitmask &);

  /* Only bit i set. */
  friend Bitmask bitmask_shift(int i);

  /* Bits 0, ..., i-1 set. */
  friend Bitmask bitmask_mask(int i);
};

#if WORDS_PER_BITMASK == 1

inline bool is_zero(const Bitmask &a)
{
  return a.val == 0;
}

inline Bitmask bitmask_shift(int i)
{
  return Bitmask(((BitmaskBaseType)1) << i);
}

inline Bitmask bitmask_mask(int i)
{
  assert((i >= 0) && (i <= BITS_PER_BITMASK));

  if (i == BITS_PER_BITMASK)
     return Bitmask((BitmaskBaseType)-1);
  else
     return Bitmask((((BitmaskBaseType)1) << i) - 1);
}

inline Bitmask operator ~(const Bitmask &a)
{
  return Bitmask(~a.val);
}

inline Bitmask operator &(const Bitmask &a, const Bitmask &b)
{
  return Bitmask(a.val & b.val);
}

inline Bitmask operator |(const Bitmask &a, const Bitmask &b)
{
  return Bitmask(a.val | b.val);
}

inline bool operator ==(const Bitmask &a, const Bitmask &b)
{
  return a.val == b.val;
}

inline bool operator !=(const Bitmask &a, const Bitmask &b)
{
  return a.val != b.val;
}

inline bool operator <(const Bitmask &a, const Bitmask &b)
{
  return a.val < b.val;
}

inline bool operator >(const Bitmask &a, const Bitmask &b)
{
  return a.val > b.val;
}

#else

inline bool is_zero(const Bitmask &a)
{
  for (int i = 0; i < WORDS_PER_BITMASK; i++)
  if (a.val[i] != 0)
     return false;

  return true;
}

inline Bitmask bitmask_shift(int i)
{
  Bitmask rv;

  rv.val[i / (8 * sizeof(BitmaskBaseType))] = 
           ((BitmaskBaseType)1) << (i % (8 * sizeof(BitmaskBaseType)));

  return rv;
}

inline Bitmask bitmask_mask(int i)
{
  Bitmask rv;
  int q = i / (8 * sizeof(BitmaskBaseType));
  int r = i % (8 * sizeof(BitmaskBaseType));

  for (int j = 0; j < q; j++)
      rv.val[j] = (BitmaskBaseType)-1;

  if (q < WORDS_PER_BITMASK)
     rv.val[q] = (((BitmaskBaseType)1) << r) - 1u;

  return rv;
}

inline Bitmask operator ~(const Bitmask &a)
{
  Bitmask rv;

  for (int j = 0; j < WORDS_PER_BITMASK; j++)
      rv.val[j] = ~a.val[j];

  return rv;
}

inline Bitmask operator &(const Bitmask &a, const Bitmask &b)
{
  Bitmask rv;

  for (int j = 0; j < WORDS_PER_BITMASK; j++)
      rv.val[j] = a.val[j] & b.val[j];

  return rv;
}

inline Bitmask operator |(const Bitmask &a, const Bitmask &b)
{
  Bitmask rv;

  for (int j = 0; j < WORDS_PER_BITMASK; j++)
      rv.val[j] = a.val[j] | b.val[j];

  return rv;
}

inline bool operator ==(const Bitmask &a, const Bitmask &b)
{
  for (int j = 0; j < WORDS_PER_BITMASK; j++)
  if (a.val[j] != b.val[j])
     return false;

  return true;
}

inline bool operator !=(const Bitmask &a, const Bitmask &b)
{
  for (int j = 0; j < WORDS_PER_BITMASK; j++)
  if (a.val[j] != b.val[j])
     return true;

  return false;
}

inline bool operator <(const Bitmask &a, const Bitmask &b)
{
  for (int j = 0; j < WORDS_PER_BITMASK; j++)
  {
    if (a.val[j] < b.val[j])
       return true;
    if (a.val[j] > b.val[j])
       return false;
  }

  return false;
}

inline bool operator >(const Bitmask &a, const Bitmask &b)
{
  for (int j = 0; j < WORDS_PER_BITMASK; j++)
  {
    if (a.val[j] < b.val[j])
       return false;
    if (a.val[j] > b.val[j])
       return true;
  }

  return false;
}

#endif

/*      E N D      B I T M A S K S                          */

/*    S H A P E                       U T I L I T I E S                 */

/* Adds expanded shape c + [cube at x] to the set, if possible. */
void expand(SymmetryGroup sg, const Shape &c, const int *x, std::set<Shape> *p_set)
{
  Shape new_shape = c;

  if (!new_shape.add_cube(x))
     return;

  switch (sg)
  {
    case SG_none:
           (void)p_set->insert(new_shape);
           break;

    case SG_trans:
           new_shape.normalize_by_translation();
           (void)p_set->insert(new_shape);
           break;

    case SG_rot:
    case SG_rot_refl:
           {
             Shape new_shape_symm;
             (void)new_shape_symm.set_to_minimal_normalized_symmetry_image(sg, new_shape);
             (void)p_set->insert(new_shape_symm);
           } 
           break;
  }
}

/* Adds shapes obtained by adding one cube to the shape c to the set. */
void expand_all(SymmetryGroup sg, const Shape &c, std::set<Shape> *p_set)
{
  int n_dims = c.get_n_dims();
  std::vector<int> x(n_dims, 0);

  for (std::size_t i = 0; i < c.n_cubes(); i++)
  {
    const int *cube = c.get_cube(i);

    x.assign(cube, cube + n_dims);

    for (int j = 0; j < n_dims; j++)
    {
      x[j]--;
      expand(sg, c, &x[0], p_set);
      x[j] += 2;
      expand(sg, c, &x[0], p_set);
      x[j]--;
    }
  }
}

/* Adds shapes obtained by expanding shapes in set_in into *p_set_out. */
void expand_set(SymmetryGroup sg, const std::set<Shape> set_in, std::set<Shape> *p_set_out)
{
  for (std::set<Shape>::const_iterator i = set_in.begin(); i != set_in.end(); ++i)
      expand_all(sg, *i, p_set_out);
}

/* Take all polycubes of dimension n_dims and n cubes, minimalize wrt symmetry group sg, */
/* and place them into the set *p_shapes. */
void get_polycubes(SymmetryGroup sg, int n_dims, int n, std::set<Shape> *p_shapes)
{
  static std::map<std::pair<SymmetryGroup, int>, std::vector<std::set<Shape> > > polycubes_found;

  if (p_shapes == NULL)
     return;

  assert((n >= 0) && (n_dims >= 0));

  if (n == 0)
  {
    p_shapes->clear();
    bool b = p_shapes->insert(Shape(n_dims)).second;
    assert(b);
    return;
  }

  if (n_dims == 0)
  {
    assert(n == 1);
    Shape s(0);
    bool b = s.add_cube((int *)NULL);
    assert(b);
    p_shapes->clear();
    b = p_shapes->insert(s).second;
    assert(b);
    return;
  }

  std::pair<std::map<std::pair<SymmetryGroup, int>, std::vector<std::set<Shape> > >::iterator,
            bool>
           p = polycubes_found.insert(std::make_pair(std::make_pair(sg, n_dims), 
                                                     std::vector<std::set<Shape> >()));

  if (p.first->second.empty())
  {
    std::vector<int> zero(n_dims, 0);
    Shape pc1(n_dims);
    bool b = pc1.add_cube(&zero[0]);
    assert(b);
    std::set<Shape> initial_set;
    initial_set.insert(pc1);
    p.first->second.push_back(initial_set);
  }

  while (p.first->second.size() < (std::size_t)n)
  {
    p.first->second.push_back(std::set<Shape>());
    expand_set(sg, *(p.first->second.end() - 2), &(p.first->second.back()));
  }

  *p_shapes = p.first->second[n - 1];
}

#define SHAPE_INDICATOR_BASE 26
#define SHAPE_INDICATOR_ALPHABET "abcdefghijklmnopqrstuvwxyz"

void print_partitioned_shape(std::size_t n_bms, const Bitmask *bms, const Shape &s)
{
  int n_dims = s.get_n_dims();
  int chars_per_cube = 1;
  std::size_t n_c = s.n_cubes();

  if (n_c == 0)
  {
    std::printf("Empty %d-dimensional shape.\n\n", n_dims);
    return;
  }

  if (n_dims == 0)
  {
    assert(n_c == 1);
    std::printf("Nonempty 0-dimensional shape.\n\n");
    return;
  }

  if (n_bms > SHAPE_INDICATOR_BASE)
  {
    std::size_t i = n_bms - 1;

    while (i >= SHAPE_INDICATOR_BASE)
    {
      i /= SHAPE_INDICATOR_BASE;
      chars_per_cube++;
    }
  }
  
  std::vector<int> x(s.get_bb_lo(), s.get_bb_lo() + n_dims);

  int z_lim_lo = (n_dims >= 3) ? s.get_bb_lo()[n_dims - 3] : 0;
  int z_lim_hi = (n_dims >= 3) ? s.get_bb_hi()[n_dims - 3] : 0;

  int y_lim_lo = (n_dims >= 2) ? s.get_bb_lo()[n_dims - 2] : 0; 
  int y_lim_hi = (n_dims >= 2) ? s.get_bb_hi()[n_dims - 2] : 0; 

  int x_lim_hi = s.get_bb_hi()[n_dims - 1];
  int x_lim_lo = s.get_bb_lo()[n_dims - 1];

  int j0;

  for (j0 = 0; j0 < n_dims; j0++)
  if (x[j0] != 0)
     break;

  if (j0 != n_dims)
  {
    std::printf("@(");
    for (int j = n_dims - 1; j >= 0; j--)
    {
      std::printf("%d",x[j]);
      if (j != 0)
         std::putchar(',');
    }
    std::printf(")\n");
  }

  std::size_t i = 0;

  for (;;)
  {
    if (n_dims > 3)
    {
      std::printf("Space ");
      for (int j = n_dims - 4; j >= 0; j--)
      {
        std::printf("%d", x[j]);
        if (j != 0)
           std::putchar(',');
      }
      std::printf(":\n");
    }

    for (int iz = z_lim_lo; iz <= z_lim_hi; iz++)
    {
      if (n_dims >= 3)
         x[n_dims - 3] = iz;

      for (int iy = y_lim_lo; iy <= y_lim_hi; iy++)
      {
        if (n_dims >= 2)
           x[n_dims - 2] = iy;

        for (x[n_dims - 1] = x_lim_lo; x[n_dims - 1] <= x_lim_hi; x[n_dims - 1]++)
        {
          if (i == n_c)
          {
            if (chars_per_cube > 1)
               std::putchar('[');
            for (int k = 0; k < chars_per_cube; k++)
                std::putchar('.');
            if (chars_per_cube > 1)
               std::putchar(']');
            continue;
          }

          int j;
          const int *p = s.get_cube(i);

          for (j = 0; j < n_dims; j++)
          if (x[j] != p[j])
              break;

          if (j != n_dims)
          {
            if (chars_per_cube > 1)
               std::putchar('[');
            for (int k = 0; k < chars_per_cube; k++)
                std::putchar('.');
            if (chars_per_cube > 1)
               std::putchar(']');
            continue;
          }

          if (chars_per_cube > 1)
             std::putchar('[');

          if (n_bms == 0)
             std::putchar('#');
          else
          {
            Bitmask bm = bitmask_shift(i);

            std::size_t i_bm;

            for (i_bm = 0; i_bm < n_bms; i_bm++)
            if (!is_zero(bm & bms[i_bm]))
               break;

            assert(i_bm < n_bms);

            std::size_t shape_indicator_power = 1u;

            for (int k = 0; k < chars_per_cube - 1; k++)
                shape_indicator_power *= SHAPE_INDICATOR_BASE;

            for (; shape_indicator_power > 0u; shape_indicator_power /= SHAPE_INDICATOR_BASE)
                std::putchar(SHAPE_INDICATOR_ALPHABET[(i_bm / shape_indicator_power) 
                                                      % SHAPE_INDICATOR_BASE]);
          }

          if (chars_per_cube > 1)
             std::putchar(']');

          i++;
        }
        std::putchar('\n');
      }
      if ((n_dims >= 3) && (iz != z_lim_hi))
      {
        for (int ix = x_lim_lo; ix <= x_lim_hi; ix++)
        {
          if (chars_per_cube > 1)
             std::fputs("==", stdout);

          for (int k = 0; k < chars_per_cube; k++)
              std::putchar('=');
        }
        std::putchar('\n');
      }
    }
    std::putchar('\n');

    int j;

    for (j = n_dims - 4; j >= 0; j--)
    {
      if (x[j]++ != s.get_bb_hi()[j])
         break;
      x[j] = s.get_bb_lo()[j];
    }

    if (j < 0)
       break;
  }
  assert(i == n_c);
}

#undef SHAPE_INDICATOR_BASE
#undef SHAPE_INDICATOR_ALPHABET

void print_shape(const Shape &s)
{
  print_partitioned_shape(0, (const Bitmask *)NULL, s);
}

void print_shape_name(const Shape &s)
{
  int n_dims = s.get_n_dims();
  std::size_t n_c = s.n_cubes();

  if (n_c == 0)
  {
    switch (n_dims)
    {
      case 2:
              std::fputs("0omino", stdout);
              break;

      case 3:
              std::fputs("0cube", stdout);
              break;

      default:
              std::printf("0:%dcube", n_dims);
              break;
    }
    return;
  }

  for (std::size_t i = 0; i < n_c; i++)
  {
    const int *coords = s.get_cube(i);
    std::putchar('(');
    for (int j = n_dims - 1; j >= 0; j--)
    {
      std::printf("%d", coords[j]);
      if (j > 0)
         std::putchar(',');
    }
    std::putchar(')');
  }
}

Bitmask overall_shape_bitmask(const Shape &shape)
{
  std::size_t c = shape.n_cubes();
  assert(c <= BITS_PER_BITMASK);
  return bitmask_mask(c);
}

/* If x + part is a subset of whole, place mask for x + part in whole to *p_mask. */
/* part and whole should have the same dimension. */
/* Returns true if it was a subset, false o.w. */
bool get_translated_mask(const int *x, const Shape &part, const Shape &whole, Bitmask *p_mask)
{
  int n_dims = part.get_n_dims();
  assert(n_dims == whole.get_n_dims());

  std::size_t n_part = part.n_cubes();
  std::size_t n_whole = whole.n_cubes();
  assert(n_whole <= BITS_PER_BITMASK);
  std::size_t i_part = 0;
  std::size_t i_whole = 0;
  Bitmask bm;

  while ((i_part < n_part) && (i_whole < n_whole))
  {
    int j = 0;
    const int *p = part.get_cube(i_part);
    const int *w = whole.get_cube(i_whole);

    for (j = 0; j < n_dims; j++)
    {
      if (x[j] + p[j] < w[j])
         return false;
     
      if (x[j] + p[j] > w[j])
         break;
    }

    if (j == n_dims)
    {  
      bm = bm | bitmask_shift(i_whole);
      i_part++;
    }
 
    i_whole++;
  }

  if (i_part != n_part)
     return false;

  if (p_mask != NULL)
     *p_mask = bm;

  return true;
}

/* Get non-translated mask for part in whole. */
bool get_mask(const Shape &part, const Shape &whole, Bitmask *p_bm)
{
  if (part.get_n_dims() == 0)
     return get_translated_mask((const int *)NULL, part, whole, p_bm);
  else
  {
    std::vector<int> x(part.get_n_dims(), 0);

    return get_translated_mask(&x[0], part, whole, p_bm);
  }
}

/* Add bitmasks for all possible translations of part into whole to *p_masks. */
/* part and whole should have the same dimension. */
void add_all_translated_masks(const Shape &part, const Shape &whole, std::set<Bitmask> *p_masks)
{
  int n_dims = part.get_n_dims();
  assert(n_dims == whole.get_n_dims());

  if (part.n_cubes() == 0)
  {
    // Insert zero mask.
    (void)p_masks->insert(Bitmask());
    return;
  }

  for (int i = 0; i < n_dims; i++)
  {
    if (part.get_bb_hi()[i] - part.get_bb_lo()[i] >
        whole.get_bb_hi()[i] - whole.get_bb_lo()[i])
       return;
  }

  std::vector<int> x(n_dims, 0);

  for (int i = 0; i < n_dims; i++)
      x[i] = whole.get_bb_lo()[i] - part.get_bb_lo()[i];

  for (;;)
  {
    Bitmask bm;

    if (get_translated_mask(((n_dims == 0) ? (const int *)NULL : &x[0]), part, whole, &bm))
       (void)p_masks->insert(bm);

    int j;

    for (j = 0; j < n_dims; j++)
    {
      if (x[j]++ != whole.get_bb_hi()[j] - part.get_bb_hi()[j])
          break;
      x[j] = whole.get_bb_lo()[j] - part.get_bb_lo()[j];
    }
    if (j == n_dims)
       break;
  }
}

/* Recover masked shape from overall shape and bitmask. */
Shape get_masked_shape(const Shape &overall, Bitmask b)
{
  std::size_t c = overall.n_cubes();
  assert(c <= BITS_PER_BITMASK);
  assert(is_zero(b & (~bitmask_mask(c))));
  
  Shape rv(overall.get_n_dims());
  for (int i = 0; i < c; i++)
      if (!is_zero(b & bitmask_shift(i)))
      {
        bool succ = rv.add_cube(overall.get_cube(i));
        assert(succ);
      } 

  return rv;
}

/*    E N D               S H A P E                       U T I L I T I E S                 */

/*                           P A R S I N G                                               */

/* Tries to read a point of the form (), (D), or (D,E,F,..,G) from *s and succeeding characters, */
/* using no more than len characters. */
/* Returns number of characters read on success, -1 on failure. */
/* Places point in *p_point. */
int parse_point(const char *s, int len, std::vector<int> *p_point)
{
  if ((len-- < 1) || (*s++ != '('))
     return -1;

  if ((len != 0) && (*s == ')'))
  {
    if (p_point != NULL)
       p_point->clear();
    return 2;
  }

  int n_tot_read = 1;
  std::vector<int> point;

  while (len != 0)
  {
    int n_read = -1;
    int i;

    if ((std::sscanf(s, "%d%n", &i, &n_read) < 1) || (n_read <= 0) || (n_read >= len))
       return -1;
  
    n_tot_read += n_read + 1;
    s += n_read;
    len -= n_read + 1;
    point.insert(point.begin(), i);

    if (*s == ')')
    {
      if (p_point != NULL)
         *p_point = point;

      return n_tot_read;
    }

    if (*s++ != ',')
       return -1;
  }

  return -1;
}

/* Tries to read a vector of the form D, DxE, DxExF, ...  from *s and succeeding characters, */
/* using no more than len characters. */
/* Returns number of characters read on success, -1 on failure. */
/* Places vector in *p_vector. */
int parse_box_vector(const char *s, int len, std::vector<int> *p_vector)
{
  std::vector<int> vec;

  int n_tot_read = 0;

  while (len != 0)
  {
    int n_read = -1;
    int i;

    if ((std::sscanf(s, "%d%n", &i, &n_read) < 1)  || (n_read <= 0) || (n_read > len))
       return -1;

    n_tot_read += n_read;
    s += n_read;
    len -= n_read;
    vec.insert(vec.begin(), i);

    if ((len == 0) || (*s != 'x'))
    {
      if (p_vector != NULL)
         *p_vector = vec;

      return n_tot_read;
    }

    s++;
    len--;
    n_tot_read++;
  }

  return -1;
}

const char *shape_term_abbreviations[][2] =
{ 
  {"3L",  "(0,0)(0,1)(1,1)"},
  {"3J",  "(1,0)(0,1)(1,1)"},
  {"3L'", "(1,0)(0,1)(1,1)"},
  {"3J'", "(0,0)(0,1)(1,1)"},

  {"4L",  "(0,0)(0,1)(0,2)(1,2)"},
  {"4J'", "(0,0)(0,1)(0,2)(1,2)"},
  {"4J",  "(1,0)(1,1)(1,2)(0,2)"},
  {"4L'", "(1,0)(1,1)(1,2)(0,2)"},
  {"4T",  "(0,0)(1,0)(2,0)(1,1)"},
  {"4T'", "(0,0)(1,0)(2,0)(1,1)"},
  {"4S",  "(0,1)(1,1)(1,0)(2,0)"},
  {"4Z'", "(0,1)(1,1)(1,0)(2,0)"},
  {"4Z",  "(0,0)(1,0)(1,1)(2,1)"},
  {"4S'", "(0,0)(1,0)(1,1)(2,1)"},

  {"5F", "(1,0)(2,0)(0,1)(1,1)(1,2)"},
  {"5F'", "(1,0)(0,0)(2,1)(1,1)(1,2)"},
  {"5I", "(0,0)(0,1)(0,2)(0,3)(0,4)"},
  {"5I'", "(0,0)(0,1)(0,2)(0,3)(0,4)"},
  {"5L", "(0,0)(0,1)(0,2)(0,3)(1,3)"},
  {"5J'", "(0,0)(0,1)(0,2)(0,3)(1,3)"},
  {"5J", "(1,0)(1,1)(1,2)(1,3)(0,3)"},
  {"5L'", "(1,0)(1,1)(1,2)(1,3)(0,3)"},
  {"5N", "(1,0)(1,1)(1,2)(0,2)(0,3)"},
  {"5N'", "(0,0)(0,1)(0,2)(1,2)(1,3)"},
  {"5P", "(0,0)(1,0)(0,1)(1,1)(0,2)"},
  {"5Q'", "(0,0)(1,0)(0,1)(1,1)(0,2)"},
  {"5P'", "(0,0)(1,0)(0,1)(1,1)(1,2)"},
  {"5Q", "(0,0)(1,0)(0,1)(1,1)(1,2)"},
  {"5T", "(0,0)(1,0)(2,0)(1,1)(1,2)"},
  {"5T'", "(0,0)(1,0)(2,0)(1,1)(1,2)"},
  {"5U", "(0,0)(0,1)(1,1)(2,1)(2,0)"},
  {"5U'", "(0,0)(0,1)(1,1)(2,1)(2,0)"},
  {"5V", "(0,0)(0,1)(0,2)(1,2)(2,2)"},
  {"5V'", "(2,0)(2,1)(2,2)(1,2)(0,2)"},
  {"5W", "(0,0)(0,1)(1,1)(1,2)(2,2)"},
  {"5W'", "(2,0)(2,1)(1,1)(1,2)(0,2)"},
  {"5X", "(1,0)(0,1)(1,1)(2,1)(1,2)"},
  {"5X'", "(1,0)(0,1)(1,1)(2,1)(1,2)"},
  {"5Y", "(1,0)(1,1)(0,1)(1,2)(1,3)"},
  {"5Y'", "(0,0)(0,1)(1,1)(0,2)(0,3)"},
  {"5Z", "(0,0)(1,0)(1,1)(1,2)(2,2)"},
  {"5S'", "(0,0)(1,0)(1,1)(1,2)(2,2)"},
  {"5S", "(2,0)(1,0)(1,1)(1,2)(0,2)"},
  {"5Z'", "(2,0)(1,0)(1,1)(1,2)(0,2)"}

};

/* Parse shape class term from *s and succeeding characters, using exactly len */
/* characters.  Returns natural number of dimensions, or -1 on failure. */
/* If p_shapes is not null, raise the shapes to n_dims dimensions and minimalize wrt */
/* symmetry group sg, and then add to *p_shapes. */
int parse_shape_term(const char *s, int len, SymmetryGroup sg, int n_dims, std::set<Shape> *p_shapes)
{
  int i;
  int n_read = -1;
  int n_read_2 = -1;
  int dims = -1;

  assert(len >= 0);

  for (int j = 0; j < sizeof(shape_term_abbreviations) / sizeof(const char *[2]); j++)
  if ((len == std::strlen(shape_term_abbreviations[j][0])) &&
      (std::strncmp(shape_term_abbreviations[j][0], s, len) == 0))
     return parse_shape_term(shape_term_abbreviations[j][1], 
                             std::strlen(shape_term_abbreviations[j][1]),
                             sg, n_dims, p_shapes);
 
  if (((std::sscanf(s, "%d%n", &i, &n_read) >= 1) && (n_read == len)) ||
      ((std::sscanf(s, "%dx%n", &i, &n_read_2) >= 1) && (n_read_2 > 0) && (n_read_2 < len)))
  {
    std::vector<int> box_size;

    if (parse_box_vector(s, len, &box_size) != len)
       return -1;

    for (std::size_t i = 0; i < box_size.size(); i++)
    if (box_size[i] < 0)
       return -1;

    dims = box_size.size();

    if (p_shapes != NULL)
    {
      assert(dims <= n_dims);
      Shape minimal_s;
      (void)minimal_s.set_to_minimal_normalized_symmetry_image
                      (restrict_group(sg, n_dims, dims), Shape(box_size));
      minimal_s.pad_dimensions(n_dims);
      (void)p_shapes->insert(minimal_s);
    }
  } else if ((len > 0) && (s[0] == '('))
  {
    std::vector<std::vector<int> > points;

    do
    {
      std::vector<int> point;
      int n_read_2 = parse_point(s, len, &point);
      if (n_read_2 < 0)
         return -1;

      points.push_back(point);

      s += n_read_2;
      len -= n_read_2;
    } while (len != 0);

    assert(!points.empty());

    dims = points[0].size();

    for (std::size_t i = 1; i < points.size(); i++)
        if (points[i].size() != dims)
           return -1;

    if (p_shapes != NULL)
    {
      assert(dims <= n_dims);

      Shape s(dims);

      for (std::size_t i = 0; i < points.size(); i++)
          (void)s.add_cube(points[i].empty() ? (const int *)NULL : &points[i][0]);

      Shape minimal_s;
      (void)minimal_s.set_to_minimal_normalized_symmetry_image
                      (restrict_group(sg, n_dims, dims), s);
      minimal_s.pad_dimensions(n_dims);
      (void)p_shapes->insert(minimal_s);
    }
  } else 
  {
    if (sg == SG_none)
       return -1;    // Can't enumerate polycubes without translation invariance.

    char keyword[11];
    int n_read_3 = -1;
    int n_read_4 = -1;

    if ((std::sscanf(s, "%d%10[a-z]%n", &i, keyword, &n_read_3) >= 2) && (n_read_3 == len))
    {
      if (std::strcmp(keyword, "omino") == 0)
         dims = 2;
      else if (std::strcmp(keyword, "cube") == 0)
         dims = 3;
    } else if ((std::sscanf(s, "%d:%dcube%n", &i, &dims, &n_read_4) >= 2) && (n_read_4 == len))
      ;
    else
      return -1;

    if ((dims < 0) || (i < 0))
       return -1;

    if ((dims == 0) && (i > 1))
       return -1;

    if (p_shapes != NULL)
    {
      assert(dims <= n_dims);
      std::set<Shape> shapes;
      get_polycubes(restrict_group(sg, n_dims, dims), dims, i, &shapes);
      for (std::set<Shape>::const_iterator p = shapes.begin(); p != shapes.end(); ++p)
      {
        Shape s = *p;
        s.pad_dimensions(n_dims);
        (void)p_shapes->insert(s);
      }
    }
  } 

  return dims;
}

/* Parse shape class from *s and succeeding characters, using exactly len characters. */
/* Returns natural number of dimensions, or -1 on failure. */
/* If p_shapes is not null, raise the shapes to n_dims dimensions and minimalize wrt */
/* symmetry group sg, and then add to *p_shapes. */
int parse_shape_class(const char *s, int len, SymmetryGroup sg, int n_dims, std::set<Shape> *p_shapes)
{
  assert(len >= 0);

  if (len == 0)
     return -1;

  const char *s2 = s;
  int max_dims = 0;

  do
  {
    const char *chunk;
    int chunk_len; 
    const char *p = std::strchr(s2, ';');

    if ((p != NULL) && (p - s2 < len))
    {
      chunk = s2;
      chunk_len = p - s2;
      s2 = p + 1;
      len -= chunk_len + 1;
    } else
    {
      chunk = s2;
      chunk_len = len;
      len = 0;
      s2 = NULL;
    }

    int dims = parse_shape_term(chunk, chunk_len, sg, n_dims, p_shapes);

    if (dims < 0)
       return -1;

    if (dims > max_dims)
       max_dims = dims;
  } while (len > 0);

  return max_dims;
}

/*             E N D                         P A R S I N G                                   */

/*                       E R R O R        C O D E S                             */

enum SolutionError { 
                     E_OK, E_piece_too_big, E_too_many_pieces, E_empty_piece, 
                     E_neg_multiplicity
                   };

/*                 E N D       E R R O R        C O D E S                       */


/*                 P I E C E       M U L T I C O V E R              P R O B L E M            */

/*
 *
 * Piece multicover problem: cover overall_mask with non-overlapping bitmasks,
 * from each of the lists piece_bms[0], piece_bms[1], ..., piece_bms[n_pieces - 1];
 * there are either <= piece_multiplicity[i]  (if !exact[i]) or == piece_multiplicity[i]
 * (if exact[i]) copies of piece i.  You must have n_pieces <= MAX_PIECES.
 *
 * The piece bitmask lists have length n_piece_bms[0], ..., n_piece_bms[n_pieces - 1];
 * each bitmask in each list is assumed to be nonzero and a subset of
 * overall_mask.
 *
 * If the callback is non-null, it is called for each solution with first argument
 * a pointer to an array (of length n_pieces) with the number of copies of each bitmask 
 * used, followed by a pointer to an array of 
 * pointers to the actual bitmasks, followed by vp.  
 * There is no guarantee as to the ordering of multiple copies of the same piece.
 *
 * An alternate problem is the shifted piece multicover problem,
 * which is the same except that the bitmasks can be shifted.
 */

struct PieceMulticover
{
  Bitmask overall_mask;
  int n_pieces;
  int piece_multiplicity[MAX_PIECES];
  bool exact[MAX_PIECES];
  std::size_t n_piece_bms[MAX_PIECES];
  Bitmask *piece_bms[MAX_PIECES];
  void (*p_callback)(int *, const Bitmask *const *, void *);
  void *vp;
}; 

void free_piece_multicover(PieceMulticover *p_pmc)
{
  for (int i = 0; i < p_pmc->n_pieces; i++)
      delete[] p_pmc->piece_bms[i];
}

/* Problem setup & output */
static PieceMulticover the_piece_multicover;
static Bitmask *the_pmc_output_ptrs[MAX_PIECES];
static int the_pmc_output_counts[MAX_PIECES];
static int the_pmc_output_multiplicity[MAX_PIECES];
static int the_pmc_n_pieces_left;

/* Flags */
static bool the_pmc_2_special = false; 
static bool the_pmc_1_special = false;
static bool the_pmc_branch_bit = true;
static bool the_pmc_branch_pc = true;

/* Helper indices */
int the_pmc_n_columns;
static Bitmask the_pmc_split_index[BITS_PER_BITMASK];
static std::size_t the_pmc_split_sizes[BITS_PER_BITMASK][MAX_PIECES];
static std::size_t the_pmc_split_sizes_higher[BITS_PER_BITMASK][MAX_PIECES];
static std::size_t the_pmc_sizes[MAX_PIECES];
static Bitmask *the_pmc_split_ptrs[BITS_PER_BITMASK][MAX_PIECES];
static Bitmask *the_pmc_split_ptrs_higher[BITS_PER_BITMASK][MAX_PIECES];
static Bitmask *the_pmc_ptrs[MAX_PIECES];
static std::vector<Bitmask> the_pmc_split_bitmasks[BITS_PER_BITMASK][MAX_PIECES];
static std::vector<Bitmask> the_pmc_split_bitmasks_higher[BITS_PER_BITMASK][MAX_PIECES];
static std::vector<Bitmask> the_pmc_bitmasks[MAX_PIECES];
static std::vector<Bitmask> the_pmc_output_vector;

/* Try to fit in just 1 copy of piece i; assumes that multiplicity allows this. */
/* Piece i uses bitmasks bms[0], ..., bms[n-1]. */
void pmc_try_1_piece(Bitmask bm, int i, const Bitmask *bms, std::size_t n)
{
  if ((n != 0) && std::binary_search(bms, bms + n, bm))
  {
    if (the_piece_multicover.p_callback != NULL)
    {
      the_pmc_output_ptrs[i][the_pmc_output_counts[i]++] = bm;
      (*the_piece_multicover.p_callback)(the_pmc_output_counts, the_pmc_output_ptrs, the_piece_multicover.vp);
      the_pmc_output_counts[i]--;
    }

    solns++;
  }
}

/* Try to fit in 1 copy each of piece i, i2. */
/* Assumes that multiplicity allows this. */
/* Piece i uses bitmasks bms1[0], ..., bms1[n1-1]; piece i2 uses bms2[0], ..., bms2[n2-1]. */
void pmc_try_2_pieces(Bitmask bm, int i, int i2, 
                                const Bitmask *bms1, std::size_t n1,
                                const Bitmask *bms2, std::size_t n2)
{
  if ((n1 != 0) && (n2 != 0))
  {
    std::size_t j = 0;
    std::size_t k = n2 - 1;
    Bitmask *output_locn = &the_pmc_output_ptrs[i][the_pmc_output_counts[i]++];
    Bitmask *output_locn_2 = &the_pmc_output_ptrs[i2][the_pmc_output_counts[i2]++];
    bool avoid_double = (i == i2) && (bms1 == bms2);

    if (avoid_double)
       assert(n1 == n2);
    /*
     * Looking for j, k s.t. bms1[j] & bms2[k] == 0, bms1[j] & (~bm) == 0, bms2[k] & (~bm) == 0,
     * bms1[j] | bms2[k] == bm.
     */
    for (;;) 
    {
      if (avoid_double && (k <= j))
         break;
      if (!is_zero((~bm) & bms1[j]))
      {
        if (++j == n1)
           break;
      }
      else if (!is_zero((~bm) & bms2[k]))
      {
        if (k-- == 0)
           break;
      }
      else if (bms1[j] < (bm & (~bms2[k])))
      {
        if (++j == n1)
           break;
      }
      else if (bms1[j] > (bm & (~bms2[k])))
      {
        if (k-- == 0)
           break;
      }
      else
      {
        if (the_piece_multicover.p_callback != NULL)
        {
          *output_locn = bms1[j];
          *output_locn_2 = bms2[k];
          (*the_piece_multicover.p_callback)(the_pmc_output_counts, the_pmc_output_ptrs, the_piece_multicover.vp);
        }

        solns++;
        if (++j == n1)
           break;
        if (k-- == 0)
           break;
      }
    }
 
    the_pmc_output_counts[i]--;
    the_pmc_output_counts[i2]--;
  }
} 

/* Special case where there is just 1 piece remaining. */
/* For each j, piece j uses bitmasks bm_ptrs[j][0], ..., bm_ptrs[j][bm_sizes[j]-1]. */
void algorithm_special_1_pmc(Bitmask overall_remaining_mask, 
                             const Bitmask *const bm_ptrs[MAX_PIECES],
                             const std::size_t bm_sizes[MAX_PIECES])
{
  int i;

  for (i = 0; i < the_piece_multicover.n_pieces; i++)
  if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
      break;

  assert(i < the_piece_multicover.n_pieces);
  assert(the_pmc_output_counts[i] + 1 == the_pmc_output_multiplicity[i]);

  pmc_try_1_piece(overall_remaining_mask, i, bm_ptrs[i], bm_sizes[i]);
}

/* Special case where there are just 2 pieces remaining. */
/* For each j, piece j uses bitmasks bm_ptrs_main[j][0], ..., bm_ptrs_main[j][bm_sizes_main[j]-1], 
 *                 and also bitmasks bm_ptrs_aux[j][0], ..., bm_ptrs_aux[j][bm_sizes_aux[j]-1]. 
 *
 * A pair of pieces is covered by a main and an aux bitmask, in some order.  
 * If it is possible to complete the problem by filling in just one piece, a main bitmask
 * is used.
 */
void algorithm_special_2_pmc(Bitmask overall_remaining_mask, 
                             const Bitmask *const bm_ptrs_main[MAX_PIECES],
                             const std::size_t bm_sizes_main[MAX_PIECES],
                             const Bitmask *const bm_ptrs_aux[MAX_PIECES],
                             const std::size_t bm_sizes_aux[MAX_PIECES])
{
  int i, i2;

  for (i = 0; i < the_piece_multicover.n_pieces; i++)
  if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
      break;

  assert(i < the_piece_multicover.n_pieces);

  if (the_pmc_output_counts[i] + 2 == the_pmc_output_multiplicity[i]) 
  {
    if (!the_piece_multicover.exact[i])
       pmc_try_1_piece(overall_remaining_mask, i, 
                                 bm_ptrs_main[i], bm_sizes_main[i]);

    pmc_try_2_pieces(overall_remaining_mask, i, i, 
                               bm_ptrs_main[i], bm_sizes_main[i], 
                               bm_ptrs_aux[i], bm_sizes_aux[i]);

  } else
  {
    assert(the_pmc_output_counts[i] + 1 == the_pmc_output_multiplicity[i]);

    for (i2 = i + 1; i2 < the_piece_multicover.n_pieces; i2++)
    if (the_pmc_output_counts[i2] < the_pmc_output_multiplicity[i2])
        break;

    assert(i2 < the_piece_multicover.n_pieces);

    assert(the_pmc_output_counts[i2] + 1 == the_pmc_output_multiplicity[i2]);

    if (!the_piece_multicover.exact[i2])
       pmc_try_1_piece(overall_remaining_mask, i, bm_ptrs_main[i], bm_sizes_main[i]);
 
    if (!the_piece_multicover.exact[i])
       pmc_try_1_piece(overall_remaining_mask, i2, bm_ptrs_main[i2], bm_sizes_main[i2]);

    pmc_try_2_pieces(overall_remaining_mask, i, i2, 
                               bm_ptrs_main[i], bm_sizes_main[i], 
                               bm_ptrs_aux[i2], bm_sizes_aux[i2]);

    if (bm_ptrs_main != bm_ptrs_aux)
       pmc_try_2_pieces(overall_remaining_mask, i, i2, 
                                  bm_ptrs_aux[i], bm_sizes_aux[i], 
                                  bm_ptrs_main[i2], bm_sizes_main[i2]);
  }
}

/*
 * This algorithm branches by looking at all possible masks which can cover
 * the first unfilled bit.
 */
void algorithm_first_bit_pmc_recurse(Bitmask overall_remaining_mask, int start_bit)
{
  nodes++;

  if (is_zero(overall_remaining_mask))
  {
    for (int i = 0; i < the_piece_multicover.n_pieces; i++)
    {
      assert(the_pmc_output_counts[i] <= the_pmc_output_multiplicity[i]);
      if (the_piece_multicover.exact[i] && 
          (the_pmc_output_counts[i] != the_pmc_output_multiplicity[i]))
         return;
    }

    if (the_piece_multicover.p_callback != NULL)
       (*the_piece_multicover.p_callback)(the_pmc_output_counts, the_pmc_output_ptrs, the_piece_multicover.vp);

    solns++;

    return;
  }

  if (the_pmc_n_pieces_left == 0)
     return;

  while (is_zero(the_pmc_split_index[start_bit] & overall_remaining_mask))
        start_bit++;

  if (the_pmc_2_special && (the_pmc_n_pieces_left == 2))
  {
    algorithm_special_2_pmc(overall_remaining_mask, 
                            the_pmc_split_ptrs[start_bit],
                            the_pmc_split_sizes[start_bit],
                            the_pmc_split_ptrs_higher[start_bit],
                            the_pmc_split_sizes_higher[start_bit]);
    return;
  }

  if (the_pmc_1_special && (the_pmc_n_pieces_left == 1))
  {
    algorithm_special_1_pmc(overall_remaining_mask, 
                            the_pmc_split_ptrs[start_bit],
                            the_pmc_split_sizes[start_bit]);
    return;
  }

  the_pmc_n_pieces_left--;

  for (int i = 0; i < the_piece_multicover.n_pieces; i++)
  if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
  {
    const Bitmask *p_bs = the_pmc_split_ptrs[start_bit][i];
    std::size_t n_bs = the_pmc_split_sizes[start_bit][i];
    Bitmask *output_locn = &the_pmc_output_ptrs[i][the_pmc_output_counts[i]];

    the_pmc_output_counts[i]++;

    checks += n_bs;

    Bitmask b;
    for (std::size_t j = 0; j < n_bs; j++)
    if (is_zero((b = p_bs[j]) & (~overall_remaining_mask)))
    {
      *output_locn = b;
      algorithm_first_bit_pmc_recurse(overall_remaining_mask & (~b), start_bit + 1);
    }

    --the_pmc_output_counts[i];
  }

  the_pmc_n_pieces_left++;
}

void setup_pmc(const PieceMulticover &c, bool is_smallest_branching)
{
  std::size_t j = 0;

  for (int i = 0; i < c.n_pieces; i++)
      j += c.piece_multiplicity[i];

  the_pmc_n_pieces_left = j;
  the_pmc_output_vector.resize(j);

  j = 0;

  for (int i = 0; i < c.n_pieces; i++)
  {
    the_pmc_output_counts[i] = 0;
    the_pmc_output_multiplicity[i] = c.piece_multiplicity[i];
    the_pmc_output_ptrs[i] = &the_pmc_output_vector[j];
    j += c.piece_multiplicity[i];
  }

  for (int i = 0; i < c.n_pieces; i++)
  for (std::size_t j = 0; j < c.n_piece_bms[i]; j++)
  {
    assert(!is_zero(c.piece_bms[i][j]));
    assert(is_zero(c.piece_bms[i][j] & (~c.overall_mask)));
  }

  if (is_smallest_branching)
  {
    for (int j = 0; j < c.n_pieces; j++)
    {
      the_pmc_bitmasks[j].assign(c.piece_bms[j],
                                 c.piece_bms[j] + c.n_piece_bms[j]);
      the_pmc_sizes[j] = c.n_piece_bms[j];
      if (the_pmc_bitmasks[j].empty())
         the_pmc_ptrs[j] = NULL;
      else
         the_pmc_ptrs[j] = &the_pmc_bitmasks[j][0];
    }
  }

  Bitmask this_bitmask;
  int i_out = 0;

  for (int i = 0; i < BITS_PER_BITMASK; i++)
  if (!is_zero((this_bitmask = bitmask_shift(i)) & c.overall_mask))
  {
    Bitmask lower_bits = bitmask_mask(i);

    for (int j = 0; j < c.n_pieces; j++)
    {
      the_pmc_split_bitmasks[i_out][j].clear();
 
      for (std::size_t k = 0; k < c.n_piece_bms[j]; k++)
      if ((!is_zero(this_bitmask & c.piece_bms[j][k])) &&
          (is_smallest_branching || 
           is_zero(lower_bits & c.piece_bms[j][k])))
         the_pmc_split_bitmasks[i_out][j].push_back(c.piece_bms[j][k]);

      the_pmc_split_sizes[i_out][j] = the_pmc_split_bitmasks[i_out][j].size();
      if (the_pmc_split_bitmasks[i_out][j].empty())
         the_pmc_split_ptrs[i_out][j] = NULL;
      else
         the_pmc_split_ptrs[i_out][j] = &the_pmc_split_bitmasks[i_out][j][0];

      if (the_pmc_2_special && (!is_smallest_branching))
      {
        Bitmask lower_or_this = lower_bits | this_bitmask;
  
        the_pmc_split_bitmasks_higher[i_out][j].clear();
  
        for (std::size_t k = 0; k < c.n_piece_bms[j]; k++)
        if (is_zero(lower_or_this & c.piece_bms[j][k]))
           the_pmc_split_bitmasks_higher[i_out][j].push_back(c.piece_bms[j][k]);
  
        the_pmc_split_sizes_higher[i_out][j] = the_pmc_split_bitmasks_higher[i_out][j].size();
        if (the_pmc_split_bitmasks_higher[i_out][j].empty())
           the_pmc_split_ptrs_higher[i_out][j] = NULL;
        else
           the_pmc_split_ptrs_higher[i_out][j] = &the_pmc_split_bitmasks_higher[i_out][j][0];
      }
    }
    the_pmc_split_index[i_out++] = this_bitmask;
  }

  the_pmc_n_columns = i_out;
}

void sort_bitmasks(PieceMulticover *p_pmc)
{
  for (int i = 0; i < p_pmc->n_pieces; i++)
  if (p_pmc->n_piece_bms[i] != 0)
      std::sort(p_pmc->piece_bms[i], p_pmc->piece_bms[i] + p_pmc->n_piece_bms[i]);
}

void setup_first_bit_pmc(const PieceMulticover &c)
{
  setup_pmc(c, false);
}

void algorithm_first_bit_pmc(const PieceMulticover &c)
{
  the_piece_multicover = c;
  setup_first_bit_pmc(c);

  get_times(&search_start);

  add_times(search_end, search_start, &setup_time);

  algorithm_first_bit_pmc_recurse(c.overall_mask, 0);

  get_times(&search_end);

  add_times(search_start, search_end, &search_time);
}

/*
 * This algorithm looks at all possible unfilled bits and their branching
 * factors with respect to being filled by pieces, and also all possible 
 * unplaced pieces and their branching factors with respect to being placed.
 * It then branches in the way which has the smallest branching factor.  It is
 * also possible to turn off either bit-based or piece-based branching.
 */
void algorithm_smallest_branching_pmc_recurse(Bitmask overall_remaining_mask)
{
  nodes++;

  if (is_zero(overall_remaining_mask))
  {
    for (int i = 0; i < the_piece_multicover.n_pieces; i++)
    {
      assert(the_pmc_output_counts[i] <= the_pmc_output_multiplicity[i]);
      if (the_piece_multicover.exact[i] && 
          (the_pmc_output_counts[i] != the_pmc_output_multiplicity[i]))
         return;
    }

    if (the_piece_multicover.p_callback != NULL)
       (*the_piece_multicover.p_callback)(the_pmc_output_counts, the_pmc_output_ptrs, the_piece_multicover.vp);

    solns++;

    return;
  }

  if (the_pmc_n_pieces_left == 0)
     return;

  if (the_pmc_2_special && (the_pmc_n_pieces_left == 2))
  {
    algorithm_special_2_pmc(overall_remaining_mask, 
                            the_pmc_ptrs, the_pmc_sizes, 
                            the_pmc_ptrs, the_pmc_sizes);
    return;
  }

  if (the_pmc_1_special && (the_pmc_n_pieces_left == 1))
  {
    algorithm_special_1_pmc(overall_remaining_mask, the_pmc_ptrs, the_pmc_sizes);
    return;
  }

  std::size_t min_count = (std::size_t)-1;
  int min_ii = 0;

  if (the_pmc_branch_bit)
  for (int b = 0; b < the_pmc_n_columns; b++)
  if (!is_zero(the_pmc_split_index[b] & overall_remaining_mask))
  {
    std::size_t count = 0;
   
    for (int i = 0; i < the_piece_multicover.n_pieces; i++)
    if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
    {
      const Bitmask *p_bs = the_pmc_split_ptrs[b][i];
      std::size_t n_bs = the_pmc_split_sizes[b][i];

      checks += n_bs;

      for (std::size_t j = 0; j < n_bs; j++)
      if (is_zero(p_bs[j] & (~overall_remaining_mask)))
         count++;
    }

    if (count < min_count)
    {
      min_count = count;
      min_ii = b + 1;
    }
  }

  if (the_pmc_branch_pc)
  for (int i = 0; i < the_piece_multicover.n_pieces; i++)
  if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
  {
    std::size_t count = 0;
    const Bitmask *p_bs = the_pmc_ptrs[i];
    std::size_t n_bs = the_pmc_sizes[i];

    checks += n_bs;

    for (std::size_t j = 0; j < n_bs; j++)
    if (is_zero(p_bs[j] & (~overall_remaining_mask)))
       count++;

    /* If not exact, we may not place the piece at all. */
    if (!the_piece_multicover.exact[i])
       count++;

    if (count < min_count)
    {
      min_count = count;
      min_ii = -(i + 1);
    }
  }

  if (min_ii == 0)
     return;         // Cannot branch any more

  the_pmc_n_pieces_left--;

  if (min_ii > 0)
  {
    min_ii--;

    for (int i = 0; i < the_piece_multicover.n_pieces; i++)
    if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
    {
      const Bitmask *p_bs = the_pmc_split_ptrs[min_ii][i];
      std::size_t n_bs = the_pmc_split_sizes[min_ii][i];
      Bitmask *output_locn = &the_pmc_output_ptrs[i][the_pmc_output_counts[i]];

      the_pmc_output_counts[i]++;

      Bitmask b;
      for (std::size_t j = 0; j < n_bs; j++)
      if (is_zero((b = p_bs[j]) & (~overall_remaining_mask)))
      {
        *output_locn = b;
        algorithm_smallest_branching_pmc_recurse(overall_remaining_mask & (~b));
      }

      --the_pmc_output_counts[i];
    }
  } else
  {
    min_ii = -(min_ii + 1);

    const Bitmask *p_bs = the_pmc_ptrs[min_ii];
    std::size_t n_bs = the_pmc_sizes[min_ii];
    Bitmask *output_locn = &the_pmc_output_ptrs[min_ii][the_pmc_output_counts[min_ii]];

    the_pmc_output_counts[min_ii]++;

    Bitmask b;
    for (std::size_t j = 0; j < n_bs; j++)
    if (is_zero((b = p_bs[j]) & (~overall_remaining_mask)))
    {
      *output_locn = b;
      the_pmc_sizes[min_ii] = j;
      algorithm_smallest_branching_pmc_recurse(overall_remaining_mask & (~b));
    }
    the_pmc_sizes[min_ii] = n_bs;

    --the_pmc_output_counts[min_ii];

    /* If not exact, we can choose not to place the piece at all. */
    if (!the_piece_multicover.exact[min_ii])
    {
      int extra_levels = the_pmc_output_multiplicity[min_ii] - the_pmc_output_counts[min_ii];
      assert(extra_levels >= 1);

      the_pmc_n_pieces_left -= (extra_levels - 1);
      the_pmc_output_multiplicity[min_ii] -= extra_levels;
      algorithm_smallest_branching_pmc_recurse(overall_remaining_mask);
      the_pmc_output_multiplicity[min_ii] += extra_levels;
      the_pmc_n_pieces_left += (extra_levels - 1);
    }
  }

  the_pmc_n_pieces_left++;
}

void setup_smallest_branching_pmc(const PieceMulticover &c)
{
  setup_pmc(c, true);
}

void algorithm_smallest_branching_pmc(const PieceMulticover &c, bool branch_bit, bool branch_pc)
{
  the_piece_multicover = c;
  the_pmc_branch_bit = branch_bit;
  the_pmc_branch_pc = branch_pc;
  setup_smallest_branching_pmc(c);

  get_times(&search_start);

  add_times(search_end, search_start, &setup_time);

  algorithm_smallest_branching_pmc_recurse(c.overall_mask);

  get_times(&search_end);

  add_times(search_start, search_end, &search_time);
}

/*
 * This algorithm uses the same method as smallest_branching_pmc_recurse()
 * to determine what bit or piece to branch on, but shuffles the bitmasks 
 * around in an attempt to reduce checking work.
 */
void algorithm_smallest_branching_pmc_2_recurse(Bitmask overall_remaining_mask, 
                                 const std::size_t (*split_sizes)[MAX_PIECES],
                                 const std::size_t *sizes)
{
  nodes++;

  if (is_zero(overall_remaining_mask))
  {
    for (int i = 0; i < the_piece_multicover.n_pieces; i++)
    {
      assert(the_pmc_output_counts[i] <= the_pmc_output_multiplicity[i]);
      if (the_piece_multicover.exact[i] && 
          (the_pmc_output_counts[i] != the_pmc_output_multiplicity[i]))
         return;
    }

    if (the_piece_multicover.p_callback != NULL)
       (*the_piece_multicover.p_callback)(the_pmc_output_counts, the_pmc_output_ptrs, the_piece_multicover.vp);

    solns++;

    return;
  }

  if (the_pmc_n_pieces_left == 0)
     return;
  
  if (the_pmc_2_special && (!the_pmc_branch_pc) && (the_pmc_n_pieces_left == 2))
  {
    algorithm_special_2_pmc(overall_remaining_mask,
                            the_pmc_ptrs, the_pmc_sizes,
                            the_pmc_ptrs, the_pmc_sizes);
    return;
  }

  if (the_pmc_1_special && (!the_pmc_branch_pc) && (the_pmc_n_pieces_left == 1))
  {
    algorithm_special_1_pmc(overall_remaining_mask, the_pmc_ptrs, the_pmc_sizes);
    return;
  }

  std::size_t counts[BITS_PER_BITMASK][MAX_PIECES];
  std::size_t overall_counts[MAX_PIECES];
  std::size_t min_count = (std::size_t)-1;
  int min_ii = 0;

  if (the_pmc_branch_bit) 
  for (int b = 0; b < the_pmc_n_columns; b++)
  if (!is_zero(the_pmc_split_index[b] & overall_remaining_mask))
  {
    std::size_t count = 0;
   
    for (int i = 0; i < the_piece_multicover.n_pieces; i++)
    if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
    {
      std::size_t sub_count = 0;

      Bitmask *p_bs = the_pmc_split_ptrs[b][i];
      std::size_t n_bs = split_sizes[b][i];

      checks += n_bs;

      for (std::size_t j = 0; j < n_bs; j++)
      if (is_zero(p_bs[j] & (~overall_remaining_mask)))
      {
        std::swap(p_bs[sub_count], p_bs[j]);
        sub_count++;
      }
      counts[b][i] = sub_count;
      count += sub_count;
    }

    if (count < min_count)
    {
      min_count = count;
      min_ii = b + 1;
    }
  }

  if (the_pmc_branch_pc)
  for (int i = 0; i < the_piece_multicover.n_pieces; i++)
  if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
  {
    std::size_t count = 0;
    Bitmask *p_bs = the_pmc_ptrs[i];
    std::size_t n_bs = sizes[i];

    checks += n_bs;

    for (std::size_t j = 0; j < n_bs; j++)
    if (is_zero(p_bs[j] & (~overall_remaining_mask)))
    {
      std::swap(p_bs[count], p_bs[j]);
      count++;
    }

    overall_counts[i] = count;

    /* If not exact, we may not place the piece at all. */
    if (!the_piece_multicover.exact[i])
       count++;
    
    if (count < min_count)
    {
      min_count = count;
      min_ii = -(i + 1);
    }
  }
  
  if (min_ii == 0)
     return;         // Cannot branch any more

  the_pmc_n_pieces_left--;

  if (min_ii > 0)
  {
    min_ii--;

    for (int i = 0; i < the_piece_multicover.n_pieces; i++)
    if (the_pmc_output_counts[i] < the_pmc_output_multiplicity[i])
    {
      const Bitmask *p_bs = the_pmc_split_ptrs[min_ii][i];
      std::size_t n_bs = counts[min_ii][i];
      Bitmask *output_locn = &the_pmc_output_ptrs[i][the_pmc_output_counts[i]];
  
      the_pmc_output_counts[i]++;
  
      for (std::size_t j = 0; j < n_bs; j++)
      {
        Bitmask b = p_bs[j];
  
        *output_locn = b;
        algorithm_smallest_branching_pmc_2_recurse(overall_remaining_mask & (~b), 
                                                   counts, overall_counts);
      }
  
      --the_pmc_output_counts[i];
    }
  } else
  {
    min_ii = -(min_ii + 1);

    const Bitmask *p_bs = the_pmc_ptrs[min_ii];
    std::size_t n_bs = overall_counts[min_ii];
    Bitmask *output_locn = &the_pmc_output_ptrs[min_ii][the_pmc_output_counts[min_ii]];

    the_pmc_output_counts[min_ii]++;

    for (std::size_t j = 0; j < n_bs; j++)
    {
      Bitmask b = p_bs[j];

      *output_locn = b;
      overall_counts[min_ii] = j;
      algorithm_smallest_branching_pmc_2_recurse(overall_remaining_mask & (~b),
                                                 counts, overall_counts);
    }

    --the_pmc_output_counts[min_ii];

    /* If not exact, we can choose not to place the piece at all. */
    if (!the_piece_multicover.exact[min_ii])
    {
      int extra_levels = the_pmc_output_multiplicity[min_ii] - the_pmc_output_counts[min_ii];
      assert(extra_levels >= 1);

      the_pmc_n_pieces_left -= (extra_levels - 1);
      the_pmc_output_multiplicity[min_ii] -= extra_levels;
      algorithm_smallest_branching_pmc_2_recurse(overall_remaining_mask, counts, overall_counts);
      the_pmc_output_multiplicity[min_ii] += extra_levels;
      the_pmc_n_pieces_left += (extra_levels - 1);
    }
  }

  the_pmc_n_pieces_left++;
}

void algorithm_smallest_branching_pmc_2(const PieceMulticover &c, bool branch_bit, bool branch_pc)
{
  the_piece_multicover = c;
  the_pmc_branch_bit = branch_bit;
  the_pmc_branch_pc = branch_pc;
  setup_smallest_branching_pmc(c);

  get_times(&search_start);

  add_times(search_end, search_start, &setup_time);

  algorithm_smallest_branching_pmc_2_recurse(c.overall_mask, the_pmc_split_sizes, the_pmc_sizes);

  get_times(&search_end);

  add_times(search_start, search_end, &search_time);
}


/*    E N D        P I E C E       M U L T I C O V E R              P R O B L E M            */

/*         S H A P E          M U L T I C O V E R                  P R O B L E M   */

/*
 * Similar to piece multicover problem, but with shapes.  The callback still takes
 * bitmasks.  Each piece can be operated on with the symmetry group specified.
 */
struct ShapeSpecifier
{
  std::set<Shape> shapes;
  int multiplicity;
  bool exact;
  SymmetryGroup symmetry_group;
}; 

struct ShapeMulticover
{
  std::vector<ShapeSpecifier> shape_classes;
  Shape overall_shape;
  void (*p_callback)(int *, const Bitmask *const *, void *);
  void *vp;
};


SolutionError shape_multicover_to_piece_multicover(const ShapeMulticover &in, PieceMulticover *p_out)
{
  if (in.overall_shape.n_cubes() > BITS_PER_BITMASK)
     return E_piece_too_big;

  if (in.shape_classes.size() > MAX_PIECES)
     return E_too_many_pieces;

  for (int i = 0; i < in.shape_classes.size(); i++)
  {
    if (in.shape_classes[i].multiplicity < 0)
       return E_neg_multiplicity;

    for (std::set<Shape>::const_iterator p = in.shape_classes[i].shapes.begin();
         p != in.shape_classes[i].shapes.end();
         ++p)
    if (p->n_cubes() == 0)
       return E_empty_piece;
  }

  p_out->n_pieces = in.shape_classes.size();
  p_out->overall_mask = overall_shape_bitmask(in.overall_shape);

  for (int i = 0; i < in.shape_classes.size(); i++)
  {
    p_out->piece_multiplicity[i] = in.shape_classes[i].multiplicity;
    p_out->exact[i] = in.shape_classes[i].exact;

    std::set<Bitmask> masks;

    switch (in.shape_classes[i].symmetry_group)
    {
      case SG_none:

        for (std::set<Shape>::const_iterator p = in.shape_classes[i].shapes.begin();     
             p != in.shape_classes[i].shapes.end();
             ++p)
        {
          Bitmask bm;

          if (get_mask(*p, in.overall_shape, &bm))
             (void)masks.insert(bm);
        }
        break;

      case SG_trans:

        for (std::set<Shape>::const_iterator p = in.shape_classes[i].shapes.begin();     
             p != in.shape_classes[i].shapes.end();
             ++p)
            add_all_translated_masks(*p, in.overall_shape, &masks);
        break;

      case SG_rot:
      case SG_rot_refl:

        for (std::set<Shape>::const_iterator p = in.shape_classes[i].shapes.begin();     
             p != in.shape_classes[i].shapes.end();
             ++p)
        {
          SymmetryGroup sg = in.shape_classes[i].symmetry_group;
          int n_symmetries = n_symmetries_in(sg, p->get_n_dims());

          for (int i = 0; i < n_symmetries; i++)
          {
            Shape s;
            s.set_to_normalized_symmetry_image(sg, i, *p);
            add_all_translated_masks(s, in.overall_shape, &masks);
          }
        }
        break;
    }
    p_out->n_piece_bms[i] = masks.size();
    Bitmask *bms = new Bitmask[masks.size()];
    std::size_t j = 0;
    for (std::set<Bitmask>::const_iterator p = masks.begin();
         p != masks.end();
         ++p)
        bms[j++] = *p;
    p_out->piece_bms[i] = bms;
  }
 
  p_out->p_callback = in.p_callback;
  p_out->vp = in.vp;
  return E_OK;
}

/*     E N D        S H A P E          M U L T I C O V E R                  P R O B L E M   */

enum TilingMethod { 
                    M_first_bit_pmc, 
                    M_smallest_branching_bit_pmc, M_smallest_branching_pc_pmc, 
                    M_smallest_branching_bit_pc_pmc,
                    M_smallest_branching_bit_pmc_2, M_smallest_branching_pc_pmc_2, 
                    M_smallest_branching_bit_pc_pmc_2
                  };


#define NUM_TILING_METHOD_NAMES 7


static struct 
{
  enum TilingMethod symbol;
  const char *name;
  const char *description;
} tiling_method_names[] = 
   { 
      { M_first_bit_pmc,                   "first-cube", "\
Tree search which, at each node, branches on the\n\
first uncovered square, cube or hypercube to see\n\
which shape should cover it and how this shape\n\
should be placed."
      },
      { M_smallest_branching_bit_pmc,      "smallest-cube", "\
Tree search which, at each node, looks for the\n\
uncovered square, cube or hypercube with the\n\
smallest branching factor, and then branches\n\
to see which shape should cover it and how this\n\
shape should be placed."
      },
      { M_smallest_branching_bit_pmc_2,    "smallest-cube-2", "\
An alternate implementation of the algorithm\n\
used by smallest-cube."
      },
      { M_smallest_branching_pc_pmc,       "smallest-shape", "\
Tree search which, at each node, looks for the\n\
shape class with the smallest branching factor in\n\
which not all shapes have been placed, and then\n\
branches to see which shape in this shape class,\n\
if any, should be used and how it should be placed."
      },
      { M_smallest_branching_pc_pmc_2,     "smallest-shape-2", "\
An alternate implementation of the algorithm used\n\
by smallest-shape."
      },
      { M_smallest_branching_bit_pc_pmc,   "smallest-cube-or-shape", "\
Tree search which, at each node, branches\n\
according to either smallest-cube or\n\
smallest-shape, depending on which has the\n\
smallest branching factor."
      },
      { M_smallest_branching_bit_pc_pmc_2, "smallest-cube-or-shape-2", "\
An alternate implementation of the algorithm\n\
used by smallest-cube-or-shape."
      }
   };

int default_tiling_method_index = 0;

void usage(const char *name)
{
  std::fprintf(stderr, "\
  Usage: %s [options] shape-1 [options] shape-2 ... [options] shape-n [options] shape-to-cover [options]\n\
\n\
\n\
  By default, this counts the number of shapes specified by each of the shape class\n\
  specifiers in shape-1, ..., shape-n, shape-to-cover.  All shapes are placed inside\n\
  the lowest-dimensional space which will contain any shape specified by any shape class\n\
  specifier.\n\
\n\
  If --tile, --tile-rot, or --tile-refl is specified, the program will also count the\n\
  number of ways to tile shape-to-cover with one shape of the form specified in shape-1,\n\
  one shape of the form specified in shape-2, and so on.  For this count, if shape-to-cover\n\
  does not specify a unique shape, the count is done once for each possible shape specified\n\
  by shape-to-cover.  If --tile is specified, no symmetry deduplication is done.  If\n\
  --tile-rot is specified, solutions which are obtainable from each other by rotation\n\
  and translation are considered to be the same, and only one solution in each\n\
  symmetry class is printed or counted; --tile-refl is the same, except that solutions\n\
  which are obtainable from each other by rotation, translation and reflection are\n\
  considered the same.\n\
\n\
  Overall options:\n\
\n\
    --tile               Count tilings.\n\
    --tile-rot           Count tilings, up to translation and rotation.\n\
    --tile-refl          Count tilings, up to translation, rotation and reflection.\n\
    --alg <alg>          Use with --tile to specify the algorithm used to search for tilings.\n\
                           Possible algorithms are:\n", name);

  for (int j = 0; j < NUM_TILING_METHOD_NAMES; j++)
  {
    const char *padding = "                             ";
    std::fprintf(stderr, "%s%s: ", padding, tiling_method_names[j].name);
    for (const char *p = tiling_method_names[j].description; *p != '\0'; p++)
    {
      std::putc(*p, stderr);
      if (*p == '\n')
      {
        std::fputs(padding, stderr);
        for (int k = 0; k < std::strlen(tiling_method_names[j].name); k++)
            std::putc(' ', stderr);
        std::fputs("  ", stderr);
      }
    }
    std::putc('\n', stderr);
  }
    
  std::fprintf(stderr, "\
                           The default algorithm is %s.\n\
    --special-1          Modify the tiling algorithm to special-case the search for tilings\n\
                           when there is only one shape remaining to be placed.\n\
                           Does not work with methods smallest-shape-2 or smallest-cube-or-shape-2.\n\
    --special-2          Modify the tiling algorithm to special-case the search for tilings\n\
                           when there are only two shapes remaining to be placed.\n\
                           Does not work with methods smallest-shape-2 or smallest-cube-or-shape-2.\n\
    --print              Print out tiling solutions and shapes, as well as counting.\n\
    --print-names        Print out names of shapes that can be reused as shape class specifiers.\n\
                           Also prints out names of shapes used in the tiling.\n\
    --print-placed-names When printing out tiling solutions, also print out names of\n\
                           shapes in the exact location used in the tiling.\n\
    --star               Print out one star per tiling solution found.\n\
\n\
  Per-class options (these work on the next shape class and all following shape\n\
  classes):\n\
\n\
    --nosymm             Don't translate or rotate shapes.\n\
    --norot              Allow shapes to be translated but not rotated.\n\
    --rot                Allow shapes to be translated and rotated but not reflected\n\
                           (This is the default.)\n\
    --refl               Allow shapes to be reflected as well as rotated and translated.\n\
    --D                  (D a string of digits.)  Use D copies of shape-i, instead of 1.\n\
                           Does not apply to shape-to-cover.\n\
    --same               Force all the copies used of shape-i to be the same,\n\
                           up to symmetry.  Does not apply to shape-to-cover.\n\
    --different          Allow different copies used of shape-i to be different.\n\
                           Does not apply to shape-to-cover.  (This is the default.)\n\
    --eq                 Use exactly 1 (or, if --D is specified, D) copies of \n\
                           shape-i, instead of 1 or fewer or D or fewer.\n\
                           Does not apply to shape-to-cover.  (This is the default.)\n\
    --le                 Use 1 or fewer (or, if --D is specified, D or fewer) copies of \n\
                           shape-i, instead of exactly 1 or D.\n\
                           Does not apply to shape-to-cover.\n\
\n\
  The shape class specifiers in shape-1, ..., shape-n, and shape-to-cover can be:\n\
\n\
       D                 D-long row of squares,\n\
       DxE               DxE rectangle of squares,\n\
       DxExF             DxExF cuboid,\n\
                         etc.\n\
\n\
  They can also be:\n\
\n\
       (S,T,...,U)(S',T',...,U')(S'',T'',...,U'')...\n\
\n\
                         The shape made out of squares, cubes or hypercubes\n\
                         with coordinates (S,T,...,U), (S',T',...,U'), etc.\n\
\n\
  Unless --nosymm has been specified, they can also be:\n\
\n\
       Domino            Any connected 2-dimensional shape with D squares.\n\
       Dcube             Any connected 3-dimensional shape with D cubes.\n\
       D:Ecube           Any connected E-dimensional shape with D hypercubes.\n\
\n\
     (Here, D, E, and F are strings of digits, and S, T, U, S', T', U', ... are\n\
      strings of digits with an optional sign at the start.)\n\
\n\
  Also, a sequence of shape class specifiers can be given, separated by semicolons\n\
    (e.g., 3omino;4omino for a shape that must be either a 3omino or 4omino.)\n\
\n\
  The following abbreviations for shape class specifiers are also possible:\n\
\n", tiling_method_names[default_tiling_method_index].name);

  for (int j = 0; j < sizeof(shape_term_abbreviations) / sizeof(const char *[2]); j++)
  {
    std::fprintf(stderr, "       %c%c%c               Short for %s.\n", 
                shape_term_abbreviations[j][0][0],
                shape_term_abbreviations[j][0][1],
                (shape_term_abbreviations[j][0][2] == '\0') 
                ? ' ' : shape_term_abbreviations[j][0][2],
                shape_term_abbreviations[j][1]);
  }

  std::fprintf(stderr, "\n\
\n\
\n\
   Examples of usage:\n\
\n\
  %s --refl 5F 5I 5L 5N 5P 5T 5U 5V 5W 5X 5Y 5Z --norot 6x10 --tile\n\
\n\
       Count the number of ways to place the 12 pentominoes into a 6x10 rectangle\n\
       (allowing reflections---that is, you are allowed to turn them over.)\n\
       There should be 9356 solutions.  (The number usually given is 2339, but each\n\
       of these 2339 solutions generates three others via reflections and 180-degree\n\
       rotation.)\n\
\n\
  %s 5:4cube --print\n\
\n\
       Print out all connected groups of 5 4-dimensional hypercubes (up to 4-dimensional\n\
       rotation and translation.)  There are 27 in all.\n\
\n\
Version %s.\n\
",
  name, name, VERSION);

  std::exit(1);
  return;
}

const char *error_message(SolutionError e)
{  
  static char s[1000];

  switch (e)
  {
    case E_OK: 
         return((const char *)NULL);
    case E_piece_too_big: 
         std::sprintf(s, "Shape too large --- maximum size %d squares, cubes or hypercubes", 
                         (int)BITS_PER_BITMASK);
         return s;
    case E_too_many_pieces: 
         std::sprintf(s, "Too many shape classes --- maximum is %d", (int)MAX_PIECES);
         return s;
    case E_empty_piece:
         return "Empty shape";
    case E_neg_multiplicity:
         return "Multiplicity negative";
  }
  /*NOTREACHED*/
  return (const char *)NULL;
}

Bitmask bm_test_a, bm_test_b, bm_test_c, bm_test_d;
Bitmask bm_test_e, bm_test_f, bm_test_g, bm_test_i;
Bitmask bm_test_j, bm_test_k, bm_test_l, bm_test_m;
Bitmask bm_test_n, bm_test_o, bm_test_p, bm_test_q;
int bm_test_h, bm_test_z, bm_test_y;

void bitmask_test() 
{
  bm_test_a = bm_test_b & bm_test_c;
  bm_test_d = ~bm_test_a;
  bm_test_e = bm_test_d;
  bm_test_f = bitmask_shift(10);
  bm_test_g = bitmask_shift(bm_test_z);
  bm_test_h = bm_test_g < bm_test_i;
  bm_test_h += 2 * (bm_test_j == bm_test_k);
  bm_test_h += 4 * (bm_test_j != bm_test_k);
  bm_test_h += 8 * is_zero(bm_test_l);
  bm_test_o = bm_test_m | bm_test_n;
  bm_test_p = bitmask_mask(bm_test_y);
  bm_test_q = bitmask_mask(15);
}

static int the_print_callback_n_pieces;
static const Shape *the_print_callback_p_shape;
static std::vector<SymmetryGroup> the_print_callback_symmetries;

static std::set<std::vector<std::set<Shape> > > the_print_callback_dedup_solutions;
static Count the_print_callback_dedup_soln_count;
static bool the_print_callback_dedup_hard;
static SymmetryGroup the_print_callback_dedup_sg;
static std::vector<int> the_print_callback_dedup_transformations;
static std::vector<int> the_print_callback_dedup_translations;


static bool do_star = false;
static bool do_print = false;
static bool do_print_names = false;
static bool do_print_placed_names = false;

/*
 * Functions to de-duplicate symmetry-identical solutions.
 */
std::vector<std::set<Shape> > 
            get_solution(int *n_bms, const Bitmask *const *bms)
{
  std::vector<std::set<Shape> > rv;

  for (int i = 0; i < the_print_callback_n_pieces; i++)
  {
    std::set<Shape> new_shapes;

    for (int j = 0; j < n_bms[i]; j++)
    { 
      bool ins = 
        new_shapes.insert
          (get_masked_shape(*the_print_callback_p_shape, bms[i][j]))
                  .second;
      assert(ins);
    }
    rv.push_back(new_shapes);
  }
  return rv;
}

std::vector<std::set<Shape> > symmetry_image_of_solution
           (const std::vector<std::set<Shape> > &solution,
            SymmetryGroup sg, int transformation, const int *translation)
{
  std::vector<std::set<Shape> > rv;

  for (std::vector<std::set<Shape> >::const_iterator i1 = solution.begin();
       i1 != solution.end(); ++i1)
  {
    std::set<Shape> new_shapes;

    for (std::set<Shape>::const_iterator i2 = i1->begin(); i2 != i1->end(); 
         ++i2)
    {
      Shape new_shape;
      new_shape.set_to_symmetry_image(sg, transformation, *i2);
      new_shape.translate_by(translation);
      bool ins = new_shapes.insert(new_shape).second;
      assert(ins); 
    }
    rv.push_back(new_shapes);
  }
  return rv;
}
            
bool is_minimal_solution(const std::vector<std::set<Shape> > &solution)
{
  std::size_t n_dims = the_print_callback_dedup_translations.size() / 
                       the_print_callback_dedup_transformations.size();

  for (std::size_t i = 0; i < the_print_callback_dedup_transformations.size(); i++)
  {
    std::vector<std::set<Shape> > image = 
      symmetry_image_of_solution(solution, 
                                 the_print_callback_dedup_sg,
                                 the_print_callback_dedup_transformations[i], 
                                 &the_print_callback_dedup_translations[i * n_dims]);
    if (image < solution)
       return false;
  }
  return true;
}

void add_solution_to_set(const std::vector<std::set<Shape> > &solution)
{
  // The code shouldn't generate duplicate solutions.
  assert(the_print_callback_dedup_solutions.find(solution) 
         == the_print_callback_dedup_solutions.end());
  
  if (!the_print_callback_dedup_transformations.empty())
  { 
    std::size_t n_dims = the_print_callback_dedup_translations.size() / 
                         the_print_callback_dedup_transformations.size();

    for (std::size_t i = 0; i < the_print_callback_dedup_transformations.size(); i++)
    {
      std::vector<std::set<Shape> > image = 
        symmetry_image_of_solution(solution, 
                                   the_print_callback_dedup_sg,
                                   the_print_callback_dedup_transformations[i], 
                                   &the_print_callback_dedup_translations[i * n_dims]);
      std::set<std::vector<std::set<Shape> > >::iterator p_image_0 = 
                        the_print_callback_dedup_solutions.find(image);
      if (p_image_0 != the_print_callback_dedup_solutions.end())
      {
        if (solution < image)
        {
          the_print_callback_dedup_solutions.erase(p_image_0);
          bool ins = the_print_callback_dedup_solutions.insert(solution).second;
          assert(ins);
        }
        return;
      }
    }
  }
  bool ins = the_print_callback_dedup_solutions.insert(solution).second;
  assert(ins);
}

void print_out_solution_set
     (std::set<std::vector<std::set<Shape> > > &solutions)
{
  if ((!do_star) && (!do_print) && 
      (!do_print_names) && (!do_print_placed_names))
     return;

  bool first = true;

  for (std::set<std::vector<std::set<Shape> > >::const_iterator 
       i0 = solutions.begin(); i0 != solutions.end(); ++i0)
  {
    if ((!do_print) && (!do_print_names) && (!do_print_placed_names))
    {
      std::fputs("*", stdout);
      std::fflush(stdout);
      continue;
    }
 
    if (do_star)
    {
      std::fputs("*\n", stdout);
      std::fflush(stdout);
    }

    if (first)
    {
      std::fputs("Tiling solutions:\n\n", stdout);
      first = false;
    }

    if (do_print)
    {
      Bitmask flattened_bms[BITS_PER_BITMASK];

      std::size_t j = 0;

      for (std::vector<std::set<Shape> >::const_iterator i1 = i0->begin(); i1 != i0->end(); ++i1)
      for (std::set<Shape>::const_iterator i2 = i1->begin(); i2 != i1->end(); ++i2)
      {
        assert(j < BITS_PER_BITMASK);
        bool ans = get_mask(*i2, *the_print_callback_p_shape, &flattened_bms[j++]);
        assert(ans);
      }

      print_partitioned_shape(j, flattened_bms, *the_print_callback_p_shape);
    }
  
    if (do_print_names || do_print_placed_names)
    {
      int i = 0;
      for (std::vector<std::set<Shape> >::const_iterator i1 = i0->begin(); i1 != i0->end(); ++i1, i++)
      {
        int j = 0;
        for (std::set<Shape>::const_iterator i2 = i1->begin(); i2 != i1->end(); ++i2, j++)
        {
          if (do_print_names)
          {
            std::printf("Shape class %d, shape %d: ", i, j);
            Shape minimal_s;
            (void)minimal_s.set_to_minimal_normalized_symmetry_image(the_print_callback_symmetries[i], *i2);
            print_shape_name(minimal_s);
            std::putchar('\n');
          }
          if (do_print_placed_names)
          {
            std::printf("Shape class %d, shape %d, as placed: ", i, j);
            print_shape_name(*i2);
            std::putchar('\n');
          }
        }
      }
      assert(i == the_print_callback_n_pieces);
      std::putchar('\n');
      std::putchar('\n');
    }
  }
}

/*
 * Callback for ShapeMulticover to print tiling solution and/or stars (*).
 */
void print_callback(int *n_bms, const Bitmask *const *bms, void *)
{
  if (the_print_callback_dedup_hard)
  {
    add_solution_to_set(get_solution(n_bms, bms));
    return;
  }

  if ((!the_print_callback_dedup_transformations.empty()) &&
      (!is_minimal_solution(get_solution(n_bms, bms))))
     return;

  if ((!do_print) && (!do_print_names) && (!do_print_placed_names))
  {
    the_print_callback_dedup_soln_count++;
    if (do_star)
    {
      std::fputs("*", stdout);
      std::fflush(stdout);
    }
    return;
  }
 
  if (do_star)
  {
    std::fputs("*\n", stdout);
    std::fflush(stdout);
  }

  if (the_print_callback_dedup_soln_count++ == 0)
     std::fputs("Tiling solutions:\n\n", stdout);

  if (do_print)
  {
    Bitmask flattened_bms[BITS_PER_BITMASK];
    int n_flattened_bms = 0;
    int j2 = 0;

    for (int i = 0; i < the_print_callback_n_pieces; i++)
        n_flattened_bms += n_bms[i];

    assert(n_flattened_bms <= BITS_PER_BITMASK);

    for (int i = 0; i < the_print_callback_n_pieces; i++)
    for (int j = 0; j < n_bms[i]; j++)
        flattened_bms[j2++] = bms[i][j];

    print_partitioned_shape(n_flattened_bms, flattened_bms, *the_print_callback_p_shape);
  }

  if (do_print_names || do_print_placed_names)
  {
    for (int i = 0; i < the_print_callback_n_pieces; i++)
    for (int j = 0; j < n_bms[i]; j++)
    {
      Shape s = get_masked_shape(*the_print_callback_p_shape, bms[i][j]);
      if (do_print_names)
      {
        std::printf("Shape class %d, shape %d: ", i, j);
        Shape minimal_s;
        (void)minimal_s.set_to_minimal_normalized_symmetry_image(the_print_callback_symmetries[i], s);
        print_shape_name(minimal_s);
        std::putchar('\n');
      }
      if (do_print_placed_names)
      {
        std::printf("Shape class %d, shape %d, as placed: ", i, j);
        print_shape_name(s);
        std::putchar('\n');
      }
    }
    std::putchar('\n');
    std::putchar('\n');
  }
}

/*
 * Callback for ShapeMulticover to print a star (*).
 */
void star_callback(int *, const Bitmask *const *, void *)
{
  std::putchar('*');
  std::fflush(stdout);
}
  
int main(int argc, char **argv)
{
  get_times(&search_end);

  bool do_tiling = false;
  ShapeMulticover smc;
  std::vector<bool> differents;
  std::set<Shape> last_shapes;
  int n_dims = -1;
  int n_shape_classes = 0;
  enum TilingMethod tiling_method = tiling_method_names[default_tiling_method_index].symbol;
  SymmetryGroup sg_dedup = SG_none;
  bool force_dedup = false;

  for (int pass = 0; pass < 2; pass++)
  {
    SymmetryGroup sg = SG_rot;
    int mult = 1;
    bool exact = true;
    bool different = true;

    for (int i = 1; i < argc; i++)
    if (argv[i][0] == '-')
    {
      if (std::strcmp(argv[i], "--nosymm") == 0)
         sg = SG_none;
      else if (std::strcmp(argv[i], "--norot") == 0)
         sg = SG_trans;
      else if (std::strcmp(argv[i], "--rot") == 0)
         sg = SG_rot;
      else if (std::strcmp(argv[i], "--refl") == 0)
         sg = SG_rot_refl;
      else if (std::strcmp(argv[i], "--print-names") == 0)
         do_print_names = true;
      else if (std::strcmp(argv[i], "--print-placed-names") == 0)
         do_print_placed_names = true;
      else if (std::strcmp(argv[i], "--print") == 0)
         do_print = true;
      else if (std::strcmp(argv[i], "--star") == 0)
         do_star = true;
      else if (std::strcmp(argv[i], "--special-1") == 0)
         the_pmc_1_special = true;
      else if (std::strcmp(argv[i], "--special-2") == 0)
         the_pmc_2_special = true;
      else if (std::strcmp(argv[i], "--tile") == 0)
         do_tiling = true;
      else if (std::strcmp(argv[i], "--tile-rot") == 0)
      {
        do_tiling = true;
        sg_dedup = SG_rot;
      }
      else if (std::strcmp(argv[i], "--tile-refl") == 0)
      {
        do_tiling = true;
        sg_dedup = SG_rot_refl;
      }
      // Undocumented option for debugging: force everything
      // to go through dedup mechanism.
      else if (std::strcmp(argv[i], "--force-dedup") == 0)
         force_dedup = true;
      else if (std::strcmp(argv[i], "--le") == 0)
         exact = false;
      else if (std::strcmp(argv[i], "--eq") == 0)
         exact = true;
      else if (std::strcmp(argv[i], "--same") == 0)
         different = false;
      else if (std::strcmp(argv[i], "--different") == 0)
         different = true;
      else if ((i + 1 < argc) && (std::strcmp(argv[i], "--alg") == 0))
      {
        int j;

        for (j = 0; j < NUM_TILING_METHOD_NAMES; j++)
        if (std::strcmp(argv[i+1], tiling_method_names[j].name) == 0)
        {
          tiling_method = tiling_method_names[j].symbol;
          break;
        }

        if (j == NUM_TILING_METHOD_NAMES)
        {
          std::fprintf(stderr, "Invalid tiling algorithm name: %s.\n\n", argv[i+1]);
          usage(argv[0]);
        }

        i++;
      }
      else if ((argv[i][1] == '-') && (argv[i][2] >= '0') && (argv[i][2] <= '9'))
      {
        int n_read = -1;
        int new_mult;
  
        if ((std::sscanf(argv[i], "--%d%n", &new_mult, &n_read) >= 1) &&
            (new_mult >= 0) && (n_read == std::strlen(argv[i])))
           mult = new_mult;
        else
        {
          std::fprintf(stderr, "Invalid option: %s.\n\n", argv[i]);
          usage(argv[0]);
        }
      }
      else
      {
        std::fprintf(stderr, "Invalid option: %s.\n\n", argv[i]);
        usage(argv[0]);
      }
    } else
    {
      int dims = parse_shape_class(argv[i], std::strlen(argv[i]), sg, -1, (std::set<Shape> *)NULL);
    
      if (dims < 0)
      {
        std::fprintf(stderr, "Invalid shape class specifier: %s.\n\n", argv[i]);
        usage(argv[0]);
      }
  
      if (pass == 0)
      {
        if (dims > n_dims)
           n_dims = dims;
 
        n_shape_classes++;
 
        continue;
      }
  
      std::set<Shape> shapes;
      int dims_2 = parse_shape_class(argv[i], std::strlen(argv[i]), sg, n_dims, &shapes);
      assert(dims == dims_2);
  
      std::printf("Shape class %d: %zu shape%s.\n",
                  (int)smc.shape_classes.size(),
                  shapes.size(),
                  ((shapes.size() == 1u) ? "" : "s"));
 
      if (do_print || do_print_names)
      {
        for (std::set<Shape>::const_iterator q = shapes.begin(); 
             q != shapes.end();
             ++q)
        {
          if (do_print_names)
          {
            std::fputs("Shape: ", stdout);
            print_shape_name(*q);
            std::putchar('\n');
          }
          if (do_print)
             print_shape(*q);
        }
      }
  
      if (smc.shape_classes.size() == n_shape_classes - 1)
      {
        last_shapes = shapes;
        continue;
      }

      ShapeSpecifier ss;

      ss.shapes = shapes;
      ss.exact = exact;
      ss.multiplicity = mult;
      ss.symmetry_group = sg;
      smc.shape_classes.push_back(ss);
      differents.push_back(different || (mult == 0));
      if (do_print_names || do_print_placed_names)
         the_print_callback_symmetries.push_back(sg);
    }
  }

  if (n_shape_classes == 0)
  {
    std::fputs("No shape classes specified.\n\n", stderr);
    usage(argv[0]);
  }

  if (!do_tiling)
     return 0;

  /* Try to tile */

  bool use_callback = (do_print || do_print_names || do_print_placed_names
                                || force_dedup || (sg_dedup != SG_none));
  if (use_callback)
     smc.p_callback = print_callback;
  else if (do_star)
     smc.p_callback = star_callback;
  else 
     smc.p_callback = NULL;

  smc.vp = NULL;

  for (std::set<Shape>::const_iterator p = last_shapes.begin();
       p != last_shapes.end(); 
       ++p)
  {
    search_time = Times();
    setup_time = Times();
    smc.overall_shape = *p;
   
    solns = 0;
    nodes = 0;
    checks = 0;

    ShapeMulticover smc2 = smc;

    for (int idx = 0; idx < n_shape_classes - 1; idx++)
    if ((!differents[idx]) && (!smc.shape_classes[idx].exact))
    {
      smc2.shape_classes[idx].multiplicity = 0;
      smc2.shape_classes[idx].exact = true;
    }

    the_print_callback_dedup_soln_count = 0;
    the_print_callback_dedup_sg = sg_dedup;
    the_print_callback_dedup_solutions.clear();
    the_print_callback_dedup_transformations.clear();
    the_print_callback_dedup_translations.clear();

    Shape orig = *p;
    orig.normalize_by_translation();

    for (int idx = 1; idx < n_symmetries_in(sg_dedup, n_dims); idx++)
    {
      Shape img;

      img.set_to_normalized_symmetry_image(sg_dedup, idx, *p);
      if (img == orig)
      {
        img.set_to_symmetry_image(sg_dedup, idx, *p);
        the_print_callback_dedup_transformations.push_back(idx);
        for (int i = 0; i < n_dims; i++)
            the_print_callback_dedup_translations
                 .push_back(p->get_bb_lo()[i] - img.get_bb_lo()[i]);
      }
    }

    if (force_dedup)
       the_print_callback_dedup_hard = true;
    else
    {
      the_print_callback_dedup_hard = false;
      for (int idx = 0; idx < n_shape_classes - 1; idx++)
         if (smc.shape_classes[idx].symmetry_group < sg_dedup)
             the_print_callback_dedup_hard = true;
    }

    for (;;)
    {
      std::vector<std::set<Shape>::const_iterator> same_places;

      for (int idx = 0; idx < n_shape_classes - 1; idx++)
          same_places.push_back(smc.shape_classes[idx].shapes.begin());

      for (;;)
      {
   
        for (int idx = 0; idx < n_shape_classes - 1; idx++)
        if ((!differents[idx]) && (smc2.shape_classes[idx].multiplicity > 0))
        {
          smc2.shape_classes[idx].shapes.clear();
          bool succ = smc2.shape_classes[idx].shapes.insert(*same_places[idx]).second;
          assert(succ);
        }
  
        { 
          PieceMulticover pmc;
          SolutionError err;

          if ((err = shape_multicover_to_piece_multicover(smc2, &pmc)) != E_OK)
          {
            std::fputs("Error solving tiling problem: ", stdout);
            std::fputs(error_message(err), stdout);
            std::fputs(".\n", stdout);
            return 2;
          }
      
          if (use_callback)
          {
            the_print_callback_n_pieces = smc2.shape_classes.size();
            the_print_callback_p_shape = &(*p);
          }

      
            if (the_pmc_1_special || the_pmc_2_special)
               sort_bitmasks(&pmc);
      
            switch (tiling_method)
            {
              case M_smallest_branching_bit_pmc:
                 algorithm_smallest_branching_pmc(pmc, true, false);
                 break;
        
              case M_smallest_branching_pc_pmc:
                 algorithm_smallest_branching_pmc(pmc, false, true);
                 break;
        
              case M_smallest_branching_bit_pc_pmc:
                 algorithm_smallest_branching_pmc(pmc, true, true);
                 break;
        
              case M_smallest_branching_bit_pmc_2:
                 algorithm_smallest_branching_pmc_2(pmc, true, false);
                 break;
        
              case M_smallest_branching_pc_pmc_2:
                 algorithm_smallest_branching_pmc_2(pmc, false, true);
                 break;
        
              case M_smallest_branching_bit_pc_pmc_2:
                 algorithm_smallest_branching_pmc_2(pmc, true, true);
                 break;
        
              case M_first_bit_pmc: 
                 algorithm_first_bit_pmc(pmc);
                 break;
            }
      

          free_piece_multicover(&pmc); 
        }
 
 
        int idx;
  
        for (idx = 0; idx < n_shape_classes - 1; idx++)
        if ((!differents[idx]) && (smc2.shape_classes[idx].multiplicity > 0))
        {
          if (++same_places[idx] != smc.shape_classes[idx].shapes.end())
              break;
  
          same_places[idx] = smc.shape_classes[idx].shapes.begin();
        }
        if (idx == n_shape_classes - 1)
           break;
      }

      int idx;
      for (idx = 0; idx < n_shape_classes - 1; idx++)
      if ((!differents[idx]) && (!smc.shape_classes[idx].exact))
      {
        if (++smc2.shape_classes[idx].multiplicity <= smc.shape_classes[idx].multiplicity)
           break;

        smc2.shape_classes[idx].multiplicity = 0;
      }
      if (idx == n_shape_classes - 1)
         break;
    }
  
    if (the_print_callback_dedup_hard)
    {
      print_out_solution_set(the_print_callback_dedup_solutions);
      solns = the_print_callback_dedup_solutions.size();
    } 
    else if (!the_print_callback_dedup_transformations.empty())
      solns = the_print_callback_dedup_soln_count;

    if (do_star && (!do_print) && (!do_print_names) && (!do_print_placed_names) && (solns > 0))
       std::putchar('\n');

    std::printf("%llu solution%s found, %llu node%s searched, %llu fit%s checked.\n", 
                solns, ((solns == 1u) ? "" : "s"),
                nodes, ((nodes == 1u) ? "" : "s"),
                checks, ((checks == 1u) ? "" : "s")
               );


#ifdef USE_CLOCK

    get_times(&search_start);
    add_times(search_end, search_start, &setup_time);
    search_end = search_start;
    std::printf("Setup time: %llu ticks rdtsc, %llu ns real, %llu ns cpu this process.\n",
                setup_time.rdtsc, setup_time.realtime, setup_time.cpuprocess);
    std::printf("Search time: %llu ticks rdtsc, %llu ns real, %llu ns cpu this process.\n",
                search_time.rdtsc, search_time.realtime, search_time.cpuprocess);
#endif
  }

  return 0;
}
